#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2012-2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys
from os.path import basename, join, isfile, isdir
import re


vendors = 'ARB NV NVX ATI 3DLABS SUN SGI SGIX SGIS INTEL 3DFX IBM MESA GREMEDY OML OES PGI I3D INGR MTX'.split()
vendors += 'KHR AMD APPLE EXT HP MESAX SUNX WIN'.split()

def parse_lineno(headers):
    hlines = {}
    for unused_abspath, relpath, source in headers:
        glversion = '0.0'
        lines = []
        for lineno, match in enumerate(re.finditer(r'^[^\n]*$', source, re.MULTILINE | re.DOTALL)):
            posa = match.start()
            pose = match.end()
            line = match.group()
            match = re.match(r'\s*#\s*define\s+GL_VERSION_([0-9_]*).*', line)
            if match is not None:
                glversion = match.group(1).replace('_', '.')
                if glversion in ['1.0', '1.1', '1.2']:
                    glversion = '1.3'
            lines.append((posa, pose, lineno, glversion))
        hlines[relpath] = lines
    return hlines
    
def find_lineno(lines, pos):
    for posa, pose, lineno, glversion in lines:
        assert pos >= posa
        if pos < pose:
            return lineno, glversion
    assert False
    
def convert_symbols_define(headers, code, symbols, hlines):
    for abspath, relpath, source in headers:
        yield '\n\n# from {}:\n'.format(abspath)
        yield "cdef extern from '{}':".format(relpath)
        for match in re.finditer(r'^\#define[ \t]+(GL_\w+).*?$', source, re.MULTILINE | re.DOTALL):
            sym = match.group(1)
            ssym = sym.split('_')
            if code and re.search(r'\b%s\b' % re.escape(sym), code) is None:
                continue
            if sym.upper() != sym or ssym[-1] in vendors:
                continue
            unused_lineno, glversion = find_lineno(hlines[relpath], match.start())
            if (glversion, sym) not in symbols:
                symbols.append((glversion, sym))
                print('.', end='', file=sys.stderr, flush=True)
                yield match.expand(r'    enum: \1')
    print(file=sys.stderr)
                
def convert_symbols_typedef(headers, code, symbols, hlines):
    for abspath, relpath, source in headers:
        yield '\n\n# from {}:\n'.format(abspath)
        for match in re.finditer(r'^(typedef(\s+(\w+))*);.*?$', source, re.MULTILINE | re.DOTALL):
            mg1 = match.group(1)
            mg1 = mg1.replace('khronos_float_t', 'float')
            mg1 = mg1.replace('khronos_', '')
            if not mg1.endswith(('_t', '64')) and not (mg1[-2:].isalpha() and mg1[-2:].isupper()):
                sym = match.group(3)
                unused_lineno, glversion = find_lineno(hlines[relpath], match.start())
                if (glversion, sym) not in symbols:
                    if code and re.search(r'\b%s\b' % re.escape(sym), code) is None:
                        continue
                    symbols.append((glversion, sym))
                    print('.', end='', file=sys.stderr, flush=True)
                    yield 'c'+mg1
    print(file=sys.stderr)
            
def const_types(types, code):
    yield '\n\ncdef extern from *:'
    for type in types.split():
        consttype = 'const_{0}_ptr'.format(type)
        if code and re.search(r'\b%s\b' % re.escape(consttype), code) is None:
            continue
        print('.', end='', file=sys.stderr, flush=True)
        yield '    ctypedef {0}* {1} "const {0}*"'.format(type, consttype)
    print(file=sys.stderr)
                           
def convert_symbols_functions(headers, code, symbols, hlines):
    for abspath, relpath, source in headers:
        yield '\n\n# from {} with:\n'.format(abspath)
        yield "cdef extern from '{}':".format(relpath)
        for match in re.finditer(
                r'^(?:GLAPI|GL_APICALL)([\s\w*]*?)(?:GLAPIENTRY|APIENTRY|GL_APIENTRY)([^(]*)\(([^)]*)\);(.*?)$',
                source, re.MULTILINE | re.DOTALL):
            mg2s2 = match.group(2).strip()[-2:]
            if mg2s2.isalpha() and mg2s2.isupper():
                continue
            for mgf in (match.group(1).find, match.group(3).find):
                if not mgf('64') == mgf('GLsync') == mgf('GLDEBUGPROC') == -1:
                    break
            else:
                if match.group(3).strip() == 'void':
                    template = r'    cdef\1\2()\4'
                else:
                    template = r'    cdef\1\2(\3)\4'
                sym = match.group(2).strip()
                unused_lineno, glversion = find_lineno(hlines[relpath], match.start())
                if (glversion, sym) not in symbols:
                    if code and re.search(r'\b%s\b' % re.escape(sym), code) is None:
                        continue
                    symbols.append((glversion, sym))
                    print('.', end='', file=sys.stderr, flush=True)
                    yield match.expand(template).replace('const ', '').replace('const*', '*') \
                                            .replace(' in,', ' in_,').replace('/*', '#/*')
    print(file=sys.stderr)
    
def convert_symbols(headers, code, print_symbols=False):
    yield '\nfrom libc.stddef cimport ptrdiff_t'
    yield 'from libc.stdint cimport int32_t, intptr_t, int8_t, uint8_t'
    
    code = '\n'.join(c for abspath, relpath, c in code)
    symbols = []
    
    hlines = parse_lineno(headers)
    
    yield from convert_symbols_define(headers, code, symbols, hlines)
    yield from convert_symbols_typedef(headers, code, symbols, hlines)
    yield from const_types('GLubyte GLboolean GLvoid GLchar GLfloat GLint GLshort '
                           'GLbyte GLuint GLushort GLclampf GLsizei GLenum void', code)
    yield from convert_symbols_functions(headers, code, symbols, hlines)
                           
    assert len(symbols) == len(set(symbols))
    assert len([s for v,s in symbols]) == len(set(s for v,s in symbols))
    symbols = sorted(symbols)
    yield '\n# GL version {} needed\n'.format(symbols[-1][0])
    
    if print_symbols:
        for glversion, sym in symbols:
            print('{}  {}'.format(glversion, sym))
                
def read_url(url, filename):
    import urllib.request
    text = urllib.request.urlopen(url).read().decode('utf-8')
    return [(url, filename, text)]
    
def read_files(path, filenames):
    code = []
    for filename in filenames:
        absfilename = join(path, filename) if isdir(path) else path
        with open(absfilename, 'rt', encoding='utf-8') as file:
            code.append((absfilename, filename, file.read()))
    return code
    
def create_pxd(headers, outfilename, code, **kwargs):
    with open(outfilename, 'wt', encoding='utf-8') as file:
        print('# {}'.format(outfilename), file=file)
        print('# generated with:', *sys.argv, file=file)
        for token in convert_symbols(headers, code, **kwargs):
            print(token, file=file)
            
def usage(status):
    print('usage:\n'
          '{0} [-h|--help]\n'
          '{0} [-s] HEADER PXDFILE [FILENAMES]\n'
          '{0} update PXDFILE\n'
          'HEADER is gl, gles2, gles3, a filename or a url.\n'
          'If FILENAMES are given, the generated gl.pxd contains only symbols found in the files.'
          .format(basename(sys.argv[0])))
    sys.exit(status)
    
def main():
    def parse_options():
        args = sys.argv[1:]
        if not args or args[0] in ['-h', '--help']:
            usage(0)
        kwargs = {}
        if args[0] == '-s':
            kwargs['print_symbols'] = True
            args = args[1:]
        return args, kwargs
    args, kwargs = parse_options()
    if len(args) < 2:
        print('wrong number of arguments')
        usage(1)
    pxdfile = args[1]
    if args[0] == 'update':
        import shlex
        args = re.search(r'^# generated with: (.*)$', read_files(pxdfile, [None])[0][-1], re.MULTILINE).group(1)
        sys.argv[1:] = shlex.split(args)[1:]
        sys.argv[2] = pxdfile
        args, unused_kwargs = parse_options()
    code = read_files('.', args[2:])
    includedir = '/usr/include'
    if args[0] == 'gl':
        headers = read_files(includedir, ['GL/gl.h', 'GL/glext.h'])
        create_pxd(headers, pxdfile, code, **kwargs)
    elif args[0] == 'gles2':
        headers = read_files(includedir, ['GLES2/gl2.h', 'GLES2/gl2ext.h'])
        create_pxd(headers, pxdfile, code, **kwargs)
    elif args[0] == 'gles3':
        headers = read_files(includedir, ['GLES3/gl3.h'])
        create_pxd(headers, pxdfile, code, **kwargs)
    elif isfile(args[0]):
        headers = read_files(args[0], ['GL/gl.h'])
        create_pxd(headers, pxdfile, code, **kwargs)
    elif args[0].startswith('http://') or args[0].startswith('https://'):
        headers = read_url(args[0], 'GL/gl.h')
        create_pxd(headers, pxdfile, code, **kwargs)
    else:
        print('wrong argument:', args[0])
        usage(1)
    

if __name__ == '__main__':
    main()
    




# -*- coding: utf-8 -*-

#  Copyright © 2013-2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from .debug import DEBUG_SHADER

N_ = lambda s: s

if DEBUG_SHADER:
    import os
    from . import config
    def shaders(names):
        for filename in os.listdir(config.SHADER_DIR):
            name = os.path.splitext(filename)[0]
            if name not in names:
                names.append(name)
        return names
else:
    shaders = lambda s: s
    
def tuple_validator(minlen, maxlen, itemtype, value, valuerange=None):
    if type(value) is not tuple:
        return False
    if not (minlen <= len(value) <= maxlen):
        return False
    for v in value:
        if type(v) is not itemtype:
            return False
        if valuerange is not None and not valuerange[0] <= v <= valuerange[1]:
            return False
    return True
    
def validate_mtype(v):
    from .model import Model
    return v in Model.cache_index['types']
    
def validate_draw_accels(v):
    if type(v) is not list: return False
    for t in v:
        if type(t) is not tuple or len(t) != 2: return False
        if type(t[0]) is not str or type(t[1]) is not str: return False
    return True
    
deprecated = type('DeprecatedType', (), {})
    
def migrate_game_size_blocks_moves_position(settings):
    mtype = settings.game.type
    size = settings.game.size
    blocks = settings.game.blocks
    moves = settings.game.moves
    position = settings.game.position
    
    if not tuple_validator(3, 100, int, size, (1, 10)):
        return
    from .model import Model
    defaultsize = Model.cache_index['type'][mtype]['defaultsize']
    if len(size) < len(defaultsize):
        return
    if len(size) > len(defaultsize):
        size = size[:len(defaultsize)]
    try:
        size = Model.cache_index['normsize'][mtype][size]
    except KeyError:
        return
    settings.games[mtype].size = size
    
    if type(blocks) is not str:
        return
    
    if type(moves) is not str:
        moves = ''
    
    if type(position) is not int or position < 0:
        position = 0
    return mtype, size, blocks, moves, position
    
def migrate_theme_face_N_color_image_mode(settings):
    mtype = settings.game.type
    facenames = ('Up', 'Down', 'Left', 'Right', 'Front', 'Back')
    for i, facename in enumerate(facenames):
        color = settings.theme.face[i].color
        image = settings.theme.face[i].image
        mode = settings.theme.face[i].mode
        if type(color) is str:  settings.theme.faces[facename].color = color
        if type(image) is str:  settings.theme.faces[facename].image = image
        if type(mode) in ['tiled', 'mosaic']:  settings.theme.faces[facename].mode = mode
        
    
schema = {
    # key:              (default,   range/enum/validator)
    #                               None: value without restriction
    #                               tuple: contains two values (min, max)
    #                               list: contains strings for the enum text,
    #                                     the index is the enum value
    #                               function: returns True, if value is valid
    'version':              (1,         lambda v: type(v) is int),
    'window.size':          ((850, 650),lambda v: tuple_validator(2, 2, int, v, (10,10000))),
    'window.divider':       (620,       (0, 10000)),
    'window.toolbar':       (True,      lambda v: type(v) is bool),
    'window.editbar':       (True,      lambda v: type(v) is bool),
    'window.statusbar':     (True,      lambda v: type(v) is bool),
    #TODO: reintroduce it, but per game.type
    #'draw.default_rotation':((-30.,39.),lambda v: tuple_validator(2, 2, float, v)),
    #TODO: deprecated
    'draw.lighting':        (True,     lambda v: type(v) is bool),
    'draw.shader':          (1,         shaders(['simple', 'lighting'])),
    'draw.selection':       (1,         ['quadrant', 'simple']),
    'draw.speed':           (30,        (1, 100)),
    'game.type':            ('Cube',    validate_mtype),
    'game.size':            (None,      deprecated),
    'game.blocks':          (None,      deprecated),
    'game.moves':           (None,      deprecated),
    'game.position':        (None,      deprecated),
    'games.*.size':         (None,      lambda v: tuple_validator(0, 10, int, v, (1, 10))),
    'theme.face.*.color':       (None,          deprecated),
    'theme.face.*.image':       (None,          deprecated),
    'theme.face.*.mode':        (None,          deprecated),
    'theme.faces.Up.color':     ('#a81407',     lambda v: type(v) is str),
    'theme.faces.Down.color':   ('#d94b1c',     lambda v: type(v) is str),
    'theme.faces.Left.color':   ('#f0c829',     lambda v: type(v) is str),
    'theme.faces.Right.color':  ('#e3e3e3',     lambda v: type(v) is str),
    'theme.faces.Front.color':  ('#1d6311',     lambda v: type(v) is str),
    'theme.faces.Back.color':   ('#00275e',     lambda v: type(v) is str),
    'theme.faces.Front_Left.color': ('#1d6311', lambda v: type(v) is str),
    'theme.faces.Front_Right.color':('#9431a8', lambda v: type(v) is str),
    'theme.faces.*.image':      ('',            lambda v: type(v) is str),
    'theme.faces.*.mode':       (0,             ['tiled', 'mosaic']),
    'theme.bgcolor':        ('#B9a177', lambda v: type(v) is str),
    'draw.accels':          ([('r', 'KP+6'), ('r-', 'Shift+KP+Right'),
                              ('l', 'KP+4'), ('l-', 'Shift+KP+Left'),
                              ('u', 'KP+8'), ('u-', 'Shift+KP+Up'),
                              ('d', 'KP+2'), ('d-', 'Shift+KP+Down'),
                              ('f', 'KP+5'), ('f-', 'Shift+KP+Clear'),
                              ('b', 'KP+0'), ('b-', 'Shift+KP+Ins'),
                              ('R', 'Ctrl+KP+8'), ('L', 'Ctrl+KP+2'),
                              ('U', 'Ctrl+KP+4'), ('D', 'Ctrl+KP+6'),
                              ('F', 'Ctrl+KP+5'), ('B', 'Ctrl+KP+0'),
                             ],         validate_draw_accels),
    'draw.zoom':            (1.4,       (0.1, 100.0)),
    # The following 6 words are for the antialiasing levels: disabled, ugly, low, medium, high, higher
    'draw.samples':         (3,         [N_('disabled'), N_('ugly'), N_('low'),
                                         N_('medium'), N_('high'), N_('higher')]),
    'draw.mirror_faces':    (False,     lambda v: type(v) is bool),
    'draw.mirror_distance': (2.1,       (0.1, 10.0)),
    'action.edit_moves':    ('Ctrl+L',  lambda v: type(v) is str),
    'action.edit_cube':     ('',        lambda v: type(v) is str),
    'action.selectmodel':   ('Ctrl+M',  lambda v: type(v) is str),
    'action.initial_state': ('',        lambda v: type(v) is str),
    'action.reset_rotation':('Ctrl+R',  lambda v: type(v) is str),
    'action.preferences':   ('Ctrl+P',  lambda v: type(v) is str),
   }




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2012-2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys, os
sys.path.insert(0, '.')
import cmd
from contextlib import suppress

from pybiklib import config, cubestate, model
from pybiklib.moves import MoveQueue
from tools import patterncube


def convert(salg):
    moves = {
        "": "",
        "F":   "f",    "B":   "b",    "R":   "r",    "L":  "l",    "U":   "u",    "D":   "d",
        "F'":  "f-",   "B'":  "b-",   "R'":  "r-",   "L'": "l-",   "U'":  "u-",   "D'":  "d-",
        "F2":  "ff",   "B2":  "bb",   "R2":  "rr",   "L2": "ll",   "U2":  "uu",   "D2":  "dd",
        "F'2": "ff",   "B'2": "bb",   "R'2": "rr",   "L'2": "ll",  "U'2": "uu",   "D'2": "dd",
        "M":   "r2-",  "E":   "u2-",  "S":   "f2",
        "M'":  "r2",   "E'":  "u2",   "S'":  "f2-",
        "M2":  "r2r2", "E2":  "u2u2", "S2":  "f2f2",
        "M2'": "r2r2", "E2'": "u2u2", "S2'": "f2f2",
        "x":   "R",   "y":   "U",   "z":   "F",
        "x'":  "R-",  "y'":  "U-",  "z'":  "F-",
        "x2":  "RR",  "y2":  "UU",  "z2":  "FF",
        "x2'": "RR",  "y2'": "UU",  "z2'": "FF",
    }
    palg = ''
    for m in salg.split(' '):
        palg += moves[m.strip('()')]
    return palg
    
def modelobj_from_string(modelstr):
    modelstr, mtype, sizes, exp = model.Model.from_string(modelstr)
    if exp is not None:
        raise ValueError('with-expression not allowed here')
    return model.Model(mtype, sizes)
    
    
class Shell (cmd.Cmd):
    intro = 'Welcome to the Pybik shell. Type help or ? to list commands.\n'
    prompt = '> '
    history_file = os.path.join(config.USER_CONFIG_DIR, 'shell-history')
    
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        model.Model.load_index()
        self.do_reset()
        
    # commands
    
    def do_reset(self, arg=None):
        '''reset to the initial state: reset'''
        self.model = modelobj_from_string('Cube 3')
        self.pattern = patterncube.PatternCube(self.model)
        
    def do_show(self, arg):
        '''show the current state: show'''
        print(self.model)
        print('pattern:', self.pattern)
        
    def do_quit(self, arg):
        '''exit Pybik shell: quit or Ctrl+D'''
        return True
        
    def do_model(self, arg):
        '''show or set current model: model [NEWMODEL]'''
        if arg:
            self.model = modelobj_from_string(arg)
        print(self.model)
        
    def do_singmaster(self, arg):
        '''convert singmaster moves to Pybik moves: singmaster NOTATION'''
        print(convert(arg))
        
    def do_posrot(self, arg):
        '''parse posrot notation of cube states: posrot NOTATION'''
        cube = cubestate.CubeState(self.model)
        if ':' not in arg:
            arg = 'idx-rot:' + arg
        cube.parse_block(arg)
        print('permutation:', [pos for pos, rot in cube._blocksn])
        print('rotations:  ', ', '.join(rot for pos, rot in cube._blocksn))
        pattern = patterncube.PatternCube(self.model)
        for pos, (idx, rot) in enumerate(cube._blocksr):
            cubie = patterncube.PatternCubie(self.model, pos)
            cubie.union(idx, rot)
            pattern[pos] = cubie
        print('pattern:', pattern)
        self.pattern = pattern
            
    def do_pattern(self, arg):
        '''parse patterns notation of cube states: pattern NOTATION'''
        pattern = patterncube.PatternCube(self.model)
        pattern.parse([c.split('=') for c in arg.split()])
        print('pattern:', pattern)
        self.pattern = pattern
        
    def do_move(self, arg):
        '''transform the current pattern: moves MOVECODE'''
        moves = MoveQueue()
        moves.parse(arg, len(arg), self.model)
        #print('moves:', moves.format(self.model)[0])
        for move in moves.moves:
            self.pattern = self.pattern.transform(move)
        print('pattern:', self.pattern)
        
    def do_match(self, arg):
        '''match the current pattern to an other pattern: match PATTERN'''
        condition = patterncube.PatternCube(self.model)
        condition.parse([c.split('=') for c in arg.split()])
        print('condition:', condition)
        accept, reject = self.pattern.fork(condition)
        print('accept:', accept)
        print('reject:', '\n        '.join(str(r) for r in reject))
        
    # ----------------
        
    def onecmd(self, line):
        cmd, arg, line = self.parseline(line)
        if not line:
            return self.emptyline()
        if cmd is None:
            return self.default(line)
        if line == 'EOF':
            print()
            return True
        self.lastcmd = line
        if cmd == '':
            return self.default(line)
        else:
            try:
                func = getattr(self, 'do_' + cmd)
            except AttributeError:
                return self.default(line)
            try:
                return func(arg)
            except Exception:
                sys.excepthook(*sys.exc_info())
                self.stdout.write('*** Error in: %s\n' % line)
                return None
                
    def preloop(self):
        with suppress(ImportError, FileNotFoundError):
            import readline
            readline.read_history_file(self.history_file)
            
    def postloop(self):
        with suppress(ImportError, OSError):
            import readline
            readline.write_history_file(self.history_file)
            
        
def main():
    Shell().cmdloop()
        
        
if __name__ == '__main__':
    main()
    




#-*- coding:utf-8 -*-

#  Copyright © 2009, 2011-2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os

from . import debug, config

N_ = lambda s: s


class Tee2(object):
    class Tee(object):
        def __init__(self, file1, file2):
            self.file1 = file1
            self.file2 = file2
        def write(self, data):
            self.file1.write(data)
            self.file2.write(data)
        def flush(self):
            self.file1.flush()
            self.file2.flush()
            
    def __init__(self, name='pybik.log', mode='w'):
        self.file = open(name, mode)
        self.stdout = sys.stdout
        self.stderr = sys.stderr
        sys.stdout = self.Tee(self.file, self.stdout)
        sys.stderr = self.Tee(self.file, self.stderr)
        
    def __del__(self):
        sys.stdout = self.stdout
        sys.stderr = self.stderr
        self.file.close()
        
        
class Options:
    debug_level_names = [__a[6:].lower() for __a in debug.__all__ if __a.startswith('DEBUG_')]
    test_arg_names = ['write-y', 'write-yes', 'write-n', 'write-no', 'write-e', 'write-error',
                      'log-widgets', 'notime']
    arg_info = [
        (None, None, '\n'.join(config.LONG_DESCRIPTION).replace('ő', 'o')), #XXX: help2man cannot handle unicode
        (None, None, 'Pybik can be fully controlled and configured via the graphical user interface.'
                     ' The options listed here are intended primarily for testing and debugging.'),
        (None, None, 'Options:'),
        (['-h', '--help'], lambda self, value: self.print_usage(),
                            'Show help message and exit'),
        (['--version'], lambda self, value: self.print_version(),
                            'Show version number and exit'),
        (['--config-file='], lambda self, value: self.__dict__.update(config_file=value),
                            'Specify the configuration filename'),
        (['--defaultconfig'], lambda self, value: self.__dict__.update(defaultconfig=True),
                            'Print default settings to stdout and exit'),
        (['--games-file='], lambda self, value: self.__dict__.update(games_file=value),
                            'Specify the filename for saved games'),
        (['--log'], lambda self, value: self.tee(),
                            'Write logfile'),
        (['--list-tests'], lambda self, value: self.print_tests(),
                            'List tests'),
        (['--test'], lambda self, value: self.select_all_tests(),
                            'Run all tests'),
        (['--test='], lambda self, value: self.__dict__.setdefault('test', []).append(value),
                            'Run test T, can be used multiple times, '
                            'use T=n where n is a number to run test T multiple times'),
        (['--test-args='], lambda self, value: self.__dict__['test_args'].extend(a for a in value.split(',') if a),
                            'T is a comma-separated list of test arguments:\n{}'
                            .format(' '.join(test_arg_names))),
        (['--debug='], lambda self, value: self.set_debug_flags(value),
                            'D is a comma-separated list of debug flags:\n{0}'
                            .format(' '.join(debug_level_names))),
        (None, None,  'Qt-Options, for a full list refer to the Qt-Reference,'
                ' but not all options make sense for this application:'),
        (['-qwindowgeometry G', '-geometry G'], None,
                            'Specifies window geometry for the main window using the X11-syntax.'
                            ' For example: 100x100+50+50'),
        (['-reverse'], None,
                            'Sets the application layout direction to left-to-right'),
        (['-widgetcount'], None,
                            'Prints debug message at the end about number of widgets left'
                            ' undestroyed and maximum number of widgets existed at the same time'),
        ]
        
    def __init__(self):
        self.config_file = config.USER_SETTINGS_FILE
        self.defaultconfig = False
        self.games_file = config.USER_GAMES_FILE
        self.pure_python = False
        self.test = []
        self.test_args = []
        self.ui_args = []
        self.systemdenable = False
        
    @staticmethod
    def format_opts(args):
        args = ((a + a.strip('--')[0].upper() if a.endswith('=') else a) for a in args)
        return '  ' + ', '.join(args)
        
    @staticmethod
    def format_help(text, indent=0):
        try:
            width = int(os.environ['COLUMNS']) - 2
        except (KeyError, ValueError):
            width = 78
        width -= indent
        def split(text):
            lines = text.split('\n')
            for line in lines:
                res = None
                words = line.split(' ')
                for word in words:
                    if res is None:
                        res = word
                    elif len(res + word) + 1 <= width:
                        res = ' '.join((res, word))
                    else:
                        yield res
                        res = word
                yield res
        return '\n'.ljust(indent+1).join(split(text))
        
    @classmethod
    def print_usage(cls):
        print('Usage:', sys.argv[0], '[options]')
        for args, unused_func, text in cls.arg_info:
            if args is None:
                print()
                print(cls.format_help(text))
            else:
                opts = cls.format_opts(args)
                if len(opts) > 18:
                    text = '\n' + text
                else:
                    opts = opts.ljust(20)
                text = cls.format_help(text, 20)
                print(opts + text)
        sys.exit(0)
        
    @staticmethod
    def print_version():
        from textwrap import fill
        print(config.APPNAME, config.VERSION)
        print()
        print(config.COPYRIGHT.replace('©', '(C)')) #XXX: help2man cannot handle unicode
        print()
        print('\n\n'.join(fill(text, width=78) for text in config.LICENSE_INFO.split('\n\n')))
        print()
        print(fill(config.LICENSE_NOT_FOUND, width=78))
        sys.exit(0)
        
    def tee(self):
        self._tee = Tee2()
        
    @staticmethod
    def print_tests():
        for filename in config.get_testdatafiles():
            print(filename)
        sys.exit(0)
        
    def select_all_tests(self):
        self.__dict__.setdefault('test', []).extend(config.get_testdatafiles())
        
    def set_debug_flags(self, value):
        if not debug.DEBUG:
            print(config.APPNAME, config.VERSION)
        debug_flags = [f for f in value.split(',') if f]
        for d in debug_flags:
            if d not in self.debug_level_names:
                print('unknown debug option:', d)
                sys.exit(1)
        debug.set_flags(debug_flags)
        print('debug flags:', *debug_flags)
        
    def parse(self, args):
        arg_functs = {o: f for opts, f, h in self.arg_info for o in opts or [] if f is not None}
        for arg in args:
            try:
                index = arg.index('=')
            except ValueError:
                value = None
            else:
                arg, value = arg[:index+1], arg[index+1:]
            try:
                func = arg_functs[arg]
            except KeyError:
                self.ui_args.append(arg)
            else:
                func(self, value)
                
        for a in self.test_args:
            if a not in self.test_arg_names:
                print('unknown test argument:', a)
                sys.exit(1)
                
def create_app(root_dir, args):
    import gettext
    from PyQt5.QtCore import QLocale, QTranslator, Qt
    from PyQt5.QtWidgets import QApplication
    from PyQt5.QtGui import QPalette, QColor
    
    # initialize QApplication
    QApplication.setAttribute(Qt.AA_X11InitThreads)
    app = QApplication(args)
    args = app.arguments()[1:]
    if args:
        print('Unknown arguments:', ' '.join(args))
        sys.exit(1)
    app.setOrganizationName(config.PACKAGE)
    app.setApplicationName(config.APPNAME)
    app.setApplicationVersion(config.VERSION)
    
    # Workaround for whatsThis-Text (white text on light background)
    palette = app.palette()
    colorfg = palette.color(QPalette.Active, QPalette.ToolTipText)
    colorbg = palette.color(QPalette.Active, QPalette.ToolTipBase)
    valuefg = colorfg.value()
    valuebg = (valuefg + 255) // 2 if valuefg < 128 else valuefg // 2
    colorbg = QColor.fromHsv(colorbg.hue(), colorbg.saturation(), valuebg)
    palette.setColor(QPalette.Active, QPalette.ToolTipBase, colorbg)
    app.setPalette(palette)
    
    # initialize translation
    language = QLocale.system().name()
    # standard Qt translation, used for e.g. standard buttons and shortcut names
    translator = QTranslator()
    translator.load('qt_' + language, config.QT_LOCALE_DIR)
    app.installTranslator(translator)
    # the rest of the app use gettext for translation
    if root_dir == sys.prefix:
        # normal installation
        LOCALEDIR = None
    else:
        # different root, e.g. /usr/local, source directory
        LOCALEDIR = config.LOCALE_DIR
    t = gettext.translation(config.PACKAGE, LOCALEDIR, languages=[language], fallback=True)
    t.install(names=['ngettext'])
    return app, translator
    
def create_window(opts):
    # create main window
    # The application module must be imported after a QApplication object is created
    from .application import MainWindow
    return MainWindow(opts)
    
    
def run(root_dir=None):
    opts = Options()
    opts.parse(sys.argv)
    
    debug.debug('Qt args:', opts.ui_args)
    
    if opts.systemdenable:
        print('systemde: nable')
        
    try:
        import PyQt5.QtCore
        import PyQt5.QtGui
        import PyQt5.QtWidgets
    except ImportError as e:
        print('This program needs PyQt5:', e)
        sys.exit(1)
    
    if opts.defaultconfig:
        # settings needs Qt
        from .settings import settings
        settings.load('')
        settings.keystore.dump(sys.stdout, all=True)
        sys.exit(0)
        
    app, translator = create_app(root_dir, opts.ui_args)
    # keep the translator object for the lifetime of the app object, so that the translation works
    if opts.test:
        from pybiktest.runner import TestRunner
        for test in TestRunner.wrap(config.TESTDATA_DIR, opts.test, opts.test_args):
            opts.config_file = test.settings_file
            opts.games_file = test.games_file
            window = create_window(opts)
            test.run(window)
            app.exec()
    else:
        window = create_window(opts)
        app.exec()
        
    if opts.systemdenable:
        print('systemdi: sable')
    




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2013-2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.



import sys, os, re
import random
from collections import namedtuple, OrderedDict
from contextlib import suppress
import tempfile
import time

from PyQt5.QtCore import Qt, QTimer
from PyQt5.QtWidgets import QAction
from PyQt5.QtTest import QTest

from . import utils
from pybiklib.settings import settings


class Quit (Exception): pass    # pylint: disable=C0321


class Logger:
    def __init__(self):
        self.lineno = None
        self.result = None
        
    def log(self, *args, **kwargs):
        if self.lineno is not None:
            args = ('line {}:'.format(self.lineno),) + args
        print(*args, **kwargs)
            
    def logf(self, format, *args):
        self.log(format.format(*args))
        
    def error(self, message, *args):
        self.log(message, *args, file=sys.stderr)
        message = message.rstrip(':')
        if message not in self.result:
            self.result[message] = 1
        else:
            self.result[message] += 1
        
    def errorf(self, message, format, *args):
        self.error(message, format.format(*args))
        
logger = Logger()
log = logger.log
logf = logger.logf
log_error = logger.error
logf_error = logger.errorf


class NamedObject:     # pylint: disable=R0903
    def __init__(self, name):
        self.name = name
    def __repr__(self):
        return self.name
matchall = NamedObject('matchall')
nomatch = NamedObject('nomatch')


def mkState(fields):
    _State = namedtuple('_State', fields)
    class State(_State):   # pylint: disable=W0232
        __slots__ = ()
        def tostr(self, other=None):
            def field_tostr(i):
                if self[i] is matchall:
                    return ''
                elif other is None or self[i] != other[i]:
                    return '{}={!r}, '.format(self.fields[i], self[i])
                else:
                    return ' ' * (len(self.fields[i]) + len(repr(self[i])) + 3)
            return ''.join([field_tostr(i) for i in range(len(self))]).rstrip(', ')
            
        fields = _State._fields
        asdict = _State._asdict
        replace = _State._replace
    State.default = State(**{f: matchall for f in State.fields})
    return State
    
    
class StateInfo:   # pylint: disable=R0903
    def __init__(self):
        self.untested_tnames = None
        self.islimit = None
        
    def __str__(self):
        return 'StateInfo: untested_t: #{}, islimit: {}'.format(
                len(self.untested_tnames), self.islimit)
                
        
class StatesInfo (dict):
    __slots__ = ()
    
    def __init__(self):
        dict.__init__(self)
        
    def istested(self, *args):
        if len(args) == 0:
            for s in self.values():
                if s.untested_tnames and s.islimit:
                    return False
            return True
        elif len(args) == 1:
            state = args[0]
            return state in self and not (self[state].untested_tnames and self[state].islimit)
        else:
            raise TypeError('istested expected at most 1 argument, got %s' % len(args))
            
    def reset(self, untested_tnames):
        for v in self.values():
            v.untested_tnames = untested_tnames[:]
            
    def get_unreached(self):
        return [s for s, si in self.items() if si.untested_tnames and si.islimit]
        
        
class Transition:
    Func = namedtuple('Func', 'name fname args func')
    
    def __init__(self, name, func=None):
        if name:
            fname, args = self._parse_name(name)
            self.funcs = [self.Func(name, fname, args, func)]
        else:
            self.funcs = []
        self.exprs_src = []
        self.exprs = {}
        self.chains = {}
        
    @property
    def name(self):
        return self.funcs[0].name if self.funcs else ''
        
    def appendfunc(self, name):
        assert self.funcs
        fname, args = self._parse_name(name)
        self.funcs.append(self.Func(name, fname, args, None))
        
    def setfunc(self, i, func):
        assert self.funcs[i].func is None
        self.funcs[i] = self.funcs[i]._replace(func=func)
        
    @staticmethod
    def _parse_name(name):
        fname, *args = name.split(' ', 1)
        args = args[0] if args else None
        return fname, args
        
        
class TestData:
    def __init__(self):
        self.constants = {}
        self.transitions = {}
        self.limits = []
        self.stateinfos = StatesInfo()
        self.path_to_untested = []
        
    def islimit(self, state):
        state = state.asdict()
        for limit in self.limits:
            try:
                if not eval(limit, self.constants, state):
                    return False
            except Exception as e:  # pylint: disable=W0703
                log_error('error in limit:', e)
        return True
        
    def reset(self):
        self.stateinfos.reset(list(self.transitions.keys()))
        
    def add_stateinfo(self, state):
        try:
            return self.stateinfos[state]
        except KeyError:
            stateinfo = dict.setdefault(self.stateinfos, state, StateInfo())
            stateinfo.untested_tnames = list(self.transitions.keys())
            stateinfo.islimit = self.islimit(state)
            return stateinfo
            
    def update_transition(self, transition, state, chain):
        self.add_stateinfo(state)
        target = chain[-1]
        if target is not None and target not in transition.chains:
            transition.chains[target] = [None] * len(transition.funcs)
            self.add_stateinfo(target)
            
    @staticmethod
    def _path_to_list(path):
        result = []
        while path is not None:
            item, path = path
            result.append(item)
        return result
        
    def get_path_to_untested(self, state):
        paths = [[(state, None), None]]
        reachable = set()
        while paths:
            path = paths.pop(0)
            state = path[0][0]
            stateinfo = self.stateinfos[state]
            reachable.add(state)
            if stateinfo.islimit and stateinfo.untested_tnames:
                return self._path_to_list(path)
            for name, transition in self.transitions.items():
                target = transition.chains[state][-1]
                if not stateinfo.islimit or self.stateinfos[target].islimit:
                    if target not in reachable:
                        paths.append([(target, name), path])
        return None
        
    def get_random_transition(self, state):
        if not self.path_to_untested:
            stateinfo = self.stateinfos[state]
            while stateinfo.untested_tnames:
                name = stateinfo.untested_tnames.pop(random.randrange(len(stateinfo.untested_tnames)))
                chain = self.transitions[name].chains.get(state, None)
                target = None if chain is None else chain[-1]
                if target is None or self.stateinfos[target].islimit:
                    return name
            self.path_to_untested = self.get_path_to_untested(state)
            if self.path_to_untested is None:
                raise Quit('End of test:')
            target, name = self.path_to_untested.pop()
            assert name is None
            assert target == state
        target, name = self.path_to_untested.pop()
        assert name is not None
        return name
        
        
class Result (OrderedDict):
    __slots__ = ('_perf_counter', '_process_time', '_showtime')
    
    def __init__(self, showtime):
        super().__init__()
        
        self._dict = dict(visited=0, states=0, transitions=0)
        self._showtime = showtime
        
    errors = property(lambda self: sum(self.values()))
    
    def __enter__(self):
        self._perf_counter = -time.perf_counter()
        self._process_time = -time.process_time()
        return self
        
    def __exit__(self, *unused_exc):
        self._perf_counter += time.perf_counter()
        self._process_time += time.process_time()
        return False
        
    @staticmethod
    def _fmt_dtime(val):
        fmt_sec = lambda val, fmt: format(val, fmt).rstrip('0').rstrip('.')
        if val < 60:
            return '{}s'.format(fmt_sec(val, '.2f'))
        s = fmt_sec(val % 60, '.1f')
        val = int(val // 60)
        if val < 60:
            return '{}m {}s'.format(val, s)
        m = val % 60
        val //= 60
        if val < 24:
            return '{}h {}m {}s'.format(val, m, s)
        h = val % 60
        val //= 24
        return '{}d {}h {}m {}s'.format(val, h, m, s)
        
    def fmt_dtime(self):
        return 'time: {}, {}'.format(self._fmt_dtime(self._perf_counter), self._fmt_dtime(self._process_time))
        
    def __str__(self):
        summary = '\n  visited: {0.visited}, states: {0.states}, transitions: {0.transitions}'.format(self)
        errors = self.errors
        if errors:
            summary += '\n  errors: {}'.format(errors)
            for k, v in self.items():
                summary += '\n    {} {}'.format(v, k)
        if self._showtime:
            summary += '\n  {}'.format(self.fmt_dtime())
        return summary
        
    def __getattr__(self, key):
        if key.startswith('_'):
            raise AttributeError()
        return self._dict[key]
        
    def __setattr__(self, key, value):
        if key.startswith('_'):
            OrderedDict.__setattr__(self, key, value)
        else:
            self._dict[key] = value
            
    def __iadd__(self, other):
        for k, v in other._dict.items():    # pylint: disable=W0212
            self._dict[k] += v
        for k, v in other.items():
            self[k] = self.get(k, 0) + v
        return self
        
    add = __iadd__
        
        
class TestReader:
    class ParseArgs:
        __slots__ = ('transition', 'initial', 'chain')
        
    def __init__(self, filename):
        self.current_testfile = filename
        self.constantstrs = []
        self.fields = OrderedDict()
        self.initial = None
        self.testdata = TestData()
        self.conditions_src = []
        self.conditions_code = []
        self.conditions_needed = []
        
    def isvalid(self, state, report=False):
        state = state.asdict()
        for i, condition in enumerate(self.conditions_code):
            try:
                if not eval(condition, self.testdata.constants, state):
                    if report:
                        logf_error('condition failed', '({}): {}', i+1, self.conditions_src[i])
                    self.conditions_needed[i] = True
                    return False
            except Exception as e:  # pylint: disable=W0703
                logf_error('error in condition', '({}): {}: {}', i+1, e, self.conditions_src[i])
        return True
        
    def _token_generic(self, pattern, line, *args):
        match = re.match(pattern, line.rstrip())
        if match is None:
            return False, args
        groups = match.groups()
        for arg, group in zip(args, groups):
            if arg is not None and arg != group:
                return False, groups
        return True, groups
        
    def token_keyvalue(self, line):
        success, (key, value) = self._token_generic(r'^(?:([\w-]+):|End\.)\s*(.*)$', line, None, None)
        return success, key, value
        
    def token_indent(self, line, indent):
        success, (indent, value) = self._token_generic(
                    r'^(\s*)(.+)$', line, indent, None)
        return success, value
        
    def token_indent2(self, line):
        success, (indent, value) = self._token_generic(
                    r'^(\s*)(\S.*)$', line, None, None)
        assert success
        return indent
        
    def token_assignvalue(self, expr):
        success, (field, value) = self._token_generic(r'^(\w+)\s*=\s*(.*)$', expr, None, None)
        return success, field, value
        
    def parse_file(self):
        structure = {
                # func:                 key,            required, repeat, multiln, children, [endhook]
                None:                   (None,          False,    False,  False,   [
                                                 self.parse_constants, self.parse_fields,
                                                 self.parse_conditions, self.parse_limits,
                                                 self.parse_initial, self.parse_transition,
                                                 self.parse_error_end,
                                                ]),
                self.parse_constants:   ('Constants',   False,    False,  True,    []),
                self.parse_fields:      ('Fields',      True,     False,  True,    [], self.parse_fields_end),
                self.parse_conditions:  ('Conditions',  False,    False,  True,    []),
                self.parse_limits:      ('Limits',      False,    False,  True,    []),
                self.parse_initial:     ('Initial',     True,     False,  True,    [self.parse_state]),
                self.parse_transition:  ('Transition',  False,    True,   True,    [self.parse_field_expr, self.parse_state]),
                self.parse_field_expr:  ('Expression',  False,    True,   False,   []),
                self.parse_state:       ('State',       False,    True,   True,    [], self.parse_state_end),
                self.parse_error_end:   (None,          False,    False,  False,   []),
        }
        valid_keys = [k for k, *unused in structure.values()]
        def _parse_indent(parentindent, line):
            indent = self.token_indent2(line)
            if parentindent is None:
                if indent != '':
                    log_error('unexpected indent')
                    return None
            else:
                if indent == parentindent:
                    return False
                if not indent.startswith(parentindent):
                    if parentindent.startswith(indent):
                        return False
                    log_error('non-matching indent')
                    return None
            return indent
        def _parse_indent_next(parentindent, line):
            indent = self.token_indent2(line)
            if parentindent == indent or parentindent.startswith(indent):
                return indent
            if indent.startswith(parentindent):
                log_error('unexpected indent')
            else:
                log_error('non-matching indent')
            return None
        def _parse_block(blocks, parentindent):
            nonlocal line
            indent = _parse_indent(parentindent, line)
            while indent is None:
                line = yield
                indent = _parse_indent(parentindent, line)
            if indent is False:
                return
            for func in blocks:
                key, required, repeat, multiline, children, *extra = structure[func]
                endhook = extra[0] if extra else lambda: None
                while True:
                    success, expr = self.token_indent(line, indent)
                    if not success:
                        break
                    success, lkey, value = self.token_keyvalue(expr)
                    if not success or lkey != key:
                        if lkey not in valid_keys:
                            log_error('unknown statement', repr(lkey))
                            line = yield
                            continue
                        if required:
                            func(None)
                            endhook()
                        break
                    if not func(value):
                        break
                    line = yield
                    if multiline:
                        while True:
                            multilineindent = _parse_indent(indent, line)
                            while multilineindent is None:
                                line = yield
                                multilineindent = _parse_indent(indent, line)
                            success, expr = self.token_indent(line, multilineindent)
                            if not success:
                                break
                            success, lkey, value = self.token_keyvalue(expr)
                            if success:
                                break
                            if not func(expr, first=False):
                                break
                            line = yield
                    yield from _parse_block(children, indent)
                    endhook()
                    if not repeat:
                        break
            assert parentindent is not None
            indent = _parse_indent_next(parentindent, line)
            while indent is None:
                line = yield
                indent = _parse_indent_next(parentindent, line)
        key, required, repeat, multiline, blocks = structure[None]
        self.parse_args = self.ParseArgs()
        try:
            line = yield
            yield from _parse_block(blocks, None)
            assert False, 'Should not be reached'
        finally:
            # GeneratorExit
            del self.parse_args
            
    def parse_constants(self, value, first=True):
        if not first:
            success, cname, cexpr = self.token_assignvalue(value)
            try:
                self.testdata.constants[cname] = eval(cexpr, self.testdata.constants)
            except SyntaxError:
                log_error('error parsing constant:', cname)
            else:
                self.constantstrs.append((cname, cexpr))
        return True
        
    def parse_fields(self, value, first=True):
        if first:
            if value is None:
                log_error('missing statement:', 'Fields')
                for f in sorted(utils.field_functions.keys()):
                    self.fields[f] = []
            return True
        else:
            success, field, fvalue = self.token_assignvalue(value)
            if not success:
                log_error('error parsing:', value)
            elif field not in list(utils.field_functions.keys()):
                log_error('unknown field:', field)
            else:
                if field in self.testdata.constants:
                    log_error('field overwrites constant:', field)
                    del self.testdata.constants[field]
                self.fields[field] = eval(fvalue, self.testdata.constants)
        return success
        
    def parse_fields_end(self):
        self.State = mkState(self.fields)
        
    def parse_conditions(self, value, first=True):
        if not first:
            self.conditions_src.append(value)
            try:
                code = compile(value, '<condition>', 'eval')
            except Exception as e:  # pylint: disable=W0703
                log_error('error in condition:', e)
                code = compile('True', '<condition>', 'eval')
            self.conditions_code.append(code)
            self.conditions_needed.append(False)
        return True
        
    def parse_limits(self, value, first=True):
        if not first:
            self.testdata.limits.append(value)
        return True
        
    def _parse_state_value(self, default, line, exprs=None):
        try:
            state = eval('dict({})'.format(line), self.testdata.constants)
        except SyntaxError:
            logf_error('error parsing state:', '{!r}', line)
            state = None
        else:
            exprs = list((exprs or {}).keys())
            for field in list(state.keys()):
                if field not in self.State.fields:
                    log_error('unselected field:', field)
                    del state[field]
                elif field in exprs:
                    logf_error('state contains expression field', '{}: {}', field, line)
            state = default and default.replace(**state)
        return state
        
    def parse_initial(self, value, first=True):
        if first:
            if value is None:
                log_error('missing statement:', 'Initial')
                value = ''
            transition = Transition(value)
            self.initial = transition
            self.parse_args.transition = transition
            self.parse_args.initial = True
        else:
            self.parse_args.transition.appendfunc(value)
        return True
        
    def parse_transition(self, value, first=True):
        if first:
            transition = Transition(value)
            if value in self.testdata.transitions:
                logf_error('duplicate transition:', '{!r}', value)
            else:
                self.testdata.transitions[value] = transition
            self.parse_args.transition = transition
            self.parse_args.initial = False
        else:
            self.parse_args.transition.appendfunc(value)
        return True
        
    def parse_field_expr(self, value):
        success, field, expr = self.token_assignvalue(value)
        if not success:
            log_error('error parsing:', value)
        else:
            transition = self.parse_args.transition
            transition.exprs_src.append((field, expr))
            if field in transition.exprs:
                logf_error('duplicate expression', 'for {}: {!r}', field, expr)
            elif field not in self.State.fields:
                log_error('unknown expression field:', field)
            else:
                try:
                    expr = compile(expr, '<expression>', 'eval')
                except Exception as e:
                    log_error('error in expression:', e)
                else:
                    transition.exprs[field] = expr
        return success
        
    def _replace_target(self, exprs, dstate, target, field, value):
        target_value = getattr(target, field)
        if field in exprs:
            try:
                value = eval(exprs[field], self.testdata.constants, dstate)
            except Exception as e:  # pylint: disable=W0703
                logf_error('error parsing expression', '{!r}: {}', exprs[field], e)
        if target_value is matchall:
            if value is matchall:
                value = dstate[field]
            target = target.replace(**{field: value})
        else:
            if field in exprs:
                if value == target_value:
                    log_error('target contains expression field:', field, 'value:', target_value)
                else:
                    log_error('target contains expression field:', field, 'value:', target_value,
                                'expected:', value)
        return target
        
    def _expand_state(self, fields, state):
        if state is None:
            return
        if len(fields) == 0:
            yield state
            return
            
        stack = [None for f in state.fields]
        i = 0
        dstate = state.asdict()
        
        while True:
            field = state.fields[i]
            value = getattr(state, field)
            values = iter(fields[field] if value is matchall else (value,))
            while True:
                try:
                    value = next(values)
                except StopIteration:
                    if i == 0:
                        return
                    i -= 1
                    values = stack[i]
                    field = state.fields[i]
                else:
                    if i == len(stack) - 1:
                        dstate[field] = value
                        yield self.State(**dstate)
                    else:
                        stack[i] = values
                        i += 1
                        dstate[field] = value
                        break
                
    def _expand_target(self, state, target, exprs):
        assert state is not None and target is not None, (state, target)
        dstate = state.asdict()
        for field, value in dstate.items():
            target = self._replace_target(exprs, dstate, target, field, value)
        return target
        
    def parse_state(self, value, first=True):
        if first:
            state = self._parse_state_value(self.State.default, value)
            self.parse_args.chain = [state]
        else:
            len_chain_max = len(self.parse_args.transition.funcs) + 1
            if len(self.parse_args.chain) == len_chain_max:
                log_error('too many states:', 'not more than', len_chain_max, 'possible')
            else:
                if value == '.':
                    value = ''
                self.parse_args.chain.append(self._parse_state_value(self.State.default, value, self.parse_args.transition.exprs))
        return True
        
    def parse_state_end(self):
        transition = self.parse_args.transition
        state, *chain = self.parse_args.chain
        while len(chain) < len(transition.funcs):
            chain.append(self.State.default)
        assert len(chain) == len(transition.funcs)
        target = chain[-1] if chain else state
        for state in self._expand_state(self.fields, state):
            if state in transition.chains and transition.chains[state][-1] is not None:
                log_error('duplicate state:', state.tostr())
            elif self.isvalid(state):
                chain_ = []
                state_ = state
                for target in chain:
                    target = self._expand_target(state_, target, transition.exprs)
                    if not self.isvalid(target, report=True):
                        log_error('invalid target state:', target.tostr())
                        target = None
                    chain_.append(target)
                    state_ = target
                transition.chains[state] = chain_
                if not self.parse_args.initial:
                    self.testdata.update_transition(transition, state, chain_)
                    
    def parse_error_end(self, value):
        if value.rstrip() != '':
            logf_error('error parsing:', '{!r}', value)
        return True     # success in case of an error just means skipping the line
        
    def read_test(self):
        try:
            parse_file = self.parse_file()
            next(parse_file)
            with open(self.current_testfile, 'rt', encoding='utf-8') as testfile:
                if self.current_testfile.endswith('.py'):
                    _locals = {}
                    exec(testfile.read(), {}, _locals)
                    testfile = _locals['lines']()
                for logger.lineno, line in enumerate(testfile):
                    if line.lstrip().startswith('#'):
                        continue
                    parse_file.send(line)
        finally:
            parse_file.send('End.')
            parse_file.close()
            logger.lineno = None
            for i, needed in enumerate(self.conditions_needed):
                if not needed:
                    logf_error('condition not needed:', '({}): {}', i+1, self.conditions_src[i])
                    
    def _ignore_fields(self, chains, igntype, exprs=None):
        result = {}
        for i, field in enumerate(self.State.fields):
            target_value = nomatch
            for state, chain in chains.items():
                target = chain[-1]
                if igntype == 'unchanged':
                    if state[i] != target[i]:
                        result = {}
                        break
                    state = state.replace(**{field: matchall})
                    target = target.replace(**{field: matchall})
                elif igntype == 'sametarget':
                    if target_value in {nomatch, target[i]}:
                        state = state.replace(**{field: matchall})
                        target_value = target[i]
                    else:
                        result = {}
                        break
                elif igntype == 'expressions':
                    if field in exprs:
                        target = target.replace(**{field: matchall})
                else:
                    assert False, igntype
                if state in result and result[state][-1] != target:
                    result = {}
                    break
                chain_ = chain[:]
                chain_[-1] = target
                result[state] = chain_
            else:
                chains = result
                result = {}
        return chains
        
    def write_test(self):
        with open(self.current_testfile, 'wt', encoding='utf-8') as testfile:
            testfile.write('Constants:\n')
            for name, value in self.constantstrs:
                print(' ', name, '=', value, file=testfile)
            testfile.write('Fields:\n')
            for field, values in self.fields.items():
                print(' ', field, '=', values, file=testfile)
            testfile.write('Conditions:\n')
            for condition in self.conditions_src:
                print(' ', condition, file=testfile)
            testfile.write('Limits:\n')
            for limit in self.testdata.limits:
                print(' ', limit, file=testfile)
            transition = self.initial
            testfile.write('Initial: %s\n' % transition.name)
            chains = list(transition.chains.items())
            #assert len(chains) == 1
            for state, chain in chains:
                testfile.write('  State: {}\n'.format(state.tostr()))
                targetstrs = []
                for target in chain:
                    targetstr = target.tostr(state)
                    if not targetstr.strip():
                        targetstr = '.'
                    targetstrs.append(targetstr)
                while targetstrs and targetstrs[-1] == '.':
                    del targetstrs[-1]
                for targetstr in targetstrs:
                    testfile.write('         {}\n'.format(targetstr))
            for name, transition in sorted(self.testdata.transitions.items()):
                testfile.write('Transition: %s\n' % name)
                chains = {s:c for s, c in transition.chains.items() if self.testdata.stateinfos[s].islimit}
                chains = self._ignore_fields(chains, 'expressions', transition.exprs)
                chains = self._ignore_fields(chains, 'unchanged')
                chains = self._ignore_fields(chains, 'sametarget')
                if len(transition.funcs) > 1:
                    for f in transition.funcs[1:]:
                        testfile.write('            {}\n'.format(f.name))
                for field, expr in sorted(transition.exprs_src):
                    testfile.write('  Expression: {} = {}\n'.format(field, expr))
                for state, chain in sorted(chains.items()):
                    testfile.write('  State: {}\n'.format(state.tostr()))
                    targetstrs = []
                    for target in chain:
                        targetstr = target.tostr(state)
                        if not targetstr.strip():
                            targetstr = '.'
                        targetstrs.append(targetstr)
                    while targetstrs and targetstrs[-1] == '.':
                        del targetstrs[-1]
                    for targetstr in targetstrs:
                        testfile.write('         {}\n'.format(targetstr))
                    
            
class TestRunner:
    _instance = None
    
    def __init__(self, testdata_dir, test, test_args):
        assert self._instance is None
        self.current_test = test
        self.write_mode = 'no'
        self.log_widgets = False
        for a in test_args:
            if a.startswith('write-n'):
                self.write_mode = 'no'
            elif a.startswith('write-y'):
                self.write_mode = 'yes'
            elif a.startswith('write-e'):
                self.write_mode = 'error'
            elif a == 'log-widgets':
                self.log_widgets = True
        filename = os.path.join(testdata_dir, test)
        self.reader = TestReader(filename)
        self.testdata = self.reader.testdata
        self.widgets = {}
        self.known_widget_functions = []
        self.field_functions = []
        
    def run(self, main_window):
        self.main_window = main_window
        self.running = True
        settings.draw.speed = 120
        QTimer.singleShot(0, self.loop)
        
    def get_state(self):
        state = {field: func() for field, func in self.field_functions}
        return self.State(**state)
        
    def find_qobjects(self, root_widget, write_template=False):
        def log_obj(indent, msg, name, obj):
            if self.log_widgets:
                logf('{}{}: {} ({})', '  '*indent, msg, name, obj.__class__.__name__)
            
        objects = [(root_widget, 0)]
        while objects:
            obj, indent = objects.pop()
            name = obj.objectName()
            if name in self.widgets:
                log_obj(indent, 'kwnobj', name, obj)
            elif name and not name.startswith('qt_'):
                self.widgets[name] = obj
                if isinstance(obj, QAction):
                    if self.reader.initial is not None:
                        transition = self.reader.initial
                        if transition.name == name:
                            transition.setfunc(0, obj.trigger)
                    if name in self.testdata.transitions:
                        transition = self.testdata.transitions[name]
                        transition.setfunc(0, obj.trigger)
                    elif write_template:
                        self.testdata.transitions[name] = Transition(name, obj.trigger)
                log_obj(indent, 'object', name, obj)
            else:
                log_obj(indent, 'intobj', name, obj)
            for child in reversed(obj.children()):
                objects.append((child, indent+1))
                
    def update_field_functions(self):
        self.field_functions = []
        mk_func = lambda wname, func: lambda: func(self.widgets[wname])
        mk_default = lambda default: lambda: default
        for field, (wname, default, func) in utils.field_functions.items():
            if field in self.State.fields:
                if wname in self.widgets:
                    self.field_functions.append((field, mk_func(wname, func)))
                else:
                    self.field_functions.append((field, mk_default(default)))
        
    def update_transition_functions(self, write_template=False):
        def mk_transition(func, widgets, *args):
            if len(widgets) == 1:
                widgets = widgets[0]
            return lambda: func(widgets, *args)
            
        def parse_transition_func(func, widgets, tfunc):
            if tfunc.args is None:
                return mk_transition(func, widgets)
            try:
                args = eval(tfunc.args, {'Qt': Qt}, self.testdata.constants)
            except Exception as e:  # pylint: disable=W0703
                logf_error('error parsing expression', '{!r}: {}', func.args, e)
                func = lambda: None
            else:
                if type(args) is not tuple:
                    args = (args,)
                func = mk_transition(func, widgets, *args)
            return func
            
        for wnames, fname, func in utils.widget_functions:
            if fname in self.known_widget_functions:
                continue
            if not all((n in self.widgets) for n in wnames):
                continue
            self.known_widget_functions.append(fname)
            widgets = [self.widgets[n] for n in wnames]
            initial = ([] if self.reader.initial is None or not self.reader.initial.funcs
                          else [(self.reader.initial.name, self.reader.initial)])
            for tname, transition in initial + list(self.testdata.transitions.items()):
                tfunc = transition.funcs[0]
                assert tname == tfunc.name
                if tfunc.fname == fname:
                    tfunc = parse_transition_func(func, widgets, tfunc)
                    transition.setfunc(0, tfunc)
                for i, chainfunc in enumerate(transition.funcs[1:]):
                    if transition.funcs[i+1].func is not None:
                        continue
                    cfname = chainfunc.fname
                    for cwnames, _cfname, cfunc in utils.widget_functions:
                        if _cfname == cfname:
                            cwidgets = [self.widgets[n] for n in cwnames]
                            cfunc = parse_transition_func(cfunc, cwidgets, chainfunc)
                            break
                    else:
                        if cfname in self.widgets and isinstance(self.widgets[cfname], QAction):
                            cwidgets = [self.widgets[cfname]]
                            cfunc = cwidgets[0].trigger
                        else:
                            assert False, '%s not found in utils.widget_functions and actions' % cfname
                    transition.setfunc(i+1, cfunc)
            if write_template:
                assert fname not in self.testdata.transitions
                self.testdata.transitions[fname] = Transition(fname, mk_transition(func, widgets))
                
    def init_test(self):
        try:
            self.reader.read_test()
        except OSError as e:
            log('Error reading test data file:', e)
        finally:
            self.State = self.reader.State
            self.fields = self.reader.fields
        assert list(self.fields.keys()) == list(self.State.fields), (list(self.fields.keys()), self.State.fields)
        write_template = False
        if not self.testdata.transitions:
            log_error('empty test')
            write_template = True
            
        # introspect ui
        self.find_qobjects(self.main_window, write_template)
        self.update_field_functions()
        self.update_transition_functions(write_template)
        
        self.current_state = self.get_state()
            
        self.testdata.reset()
        if write_template:
            raise Quit('Template created:')
        
    def check_initial(self):
        transition = self.reader.initial
        if transition is None:
            transition = self.reader.initial = Transition('')
        if len(transition.chains) > 1:
            logf_error('ambiguous initial transition:', '\n  {}: {} states',
                            transition.name, len(transition.chains))
        for state in list(transition.chains.keys()):
            if state != self.current_state:
                logf_error('wrong initial state:', '\n     found: {}\n  expected: {}',
                            self.current_state.tostr(), state)
                del transition.chains[state]
        if len(transition.chains) == 0:
            logf_error('missing initial state:', '{}', self.current_state)
            transition.chains[self.current_state] = [None for f in transition.funcs]
                    
    def check_transition(self, transition, chain, i, initial=False):
        target = chain[i]
        current_state = self.get_state()
        if target != current_state:
            if target is None:
                logf_error('unknown transition:', '\n  {}: {} -> {}',
                            transition.name, self.current_state.tostr(), current_state.tostr())
            else:
                logf_error('wrong target', 'for {}: {}\n     found: -> {}\n  expected: -> {}',
                            transition.name, self.current_state.tostr(), current_state.tostr(), target.tostr())
            chain[i] = current_state
            if not initial and i+1 == len(chain):
                self.testdata.update_transition(transition, self.current_state, chain)
        self.current_state = current_state
        
    def check_result(self):
        for name, transition in list(self.testdata.transitions.items()):
            if transition.funcs[0].func is None:
                log_error('unused transition:', name)
                del self.testdata.transitions[name]
        all_states = list(self.testdata.stateinfos.keys())
        all_transitions = [(name, state) for name, transition in self.testdata.transitions.items()
                                            for state in transition.chains.keys()]
        logger.result.states = len(all_states)    # pylint: disable=W0201
        logger.result.transitions = len(all_transitions)  # pylint: disable=W0201
        unreached = self.testdata.stateinfos.get_unreached()
        for state in unreached:
            log_error('unreached', state.tostr())
        field_values = {}
        for name, transition in self.testdata.transitions.items():
            chains = transition.chains
            for state, chain in list(chains.items()):
                target = chain[-1]
                if state in unreached or not self.testdata.stateinfos[state].islimit or target is None:
                    del chains[state]
                    continue
                for field, svalue, tvalue in zip(self.State.fields, state, target):
                    field_values.setdefault(field, set()).update([svalue, tvalue])
        for field, values in self.fields.items():
            if field in field_values:
                if set(values) != field_values[field]:
                    log_error('changed field values:', field)
                    values[:] = sorted(field_values[field])
            else:
                log_error('unused field', field)
                
    def step(self, transition, initial=False):
        chain = transition.chains.get(self.current_state, None)
        if chain is None:
            logf_error('unknown chain:', '\n  {}: {}',
                        transition.name, self.current_state.tostr())
            chain = [None] * len(transition.funcs)
            transition.chains[self.current_state] = chain
        for i, func in enumerate(transition.funcs):
            yield from self._step(func)
            self.check_transition(transition, chain, i, initial)
        
    def _step(self, func):
        postfunc = func.func()
        if postfunc == 'find_qobjects':
            self.find_qobjects(self.main_window)
            self.update_field_functions()
            self.update_transition_functions()
        logger.result.visited += 1    # pylint: disable=E1101
        yield
        while self.main_window.is_animating():
            yield
            
    def loop(self):
        if not QTest.qWaitForWindowExposed(self.main_window, timeout=5000):
            log_error('Wait for window timed out')
            self.running = False
            return
        for unused in self._iloop():
            if not self.main_window.isVisible():
                log_error('Unexpected end of test')
                self.running = False
                return
            QTest.qWait(10)
        # without this line sometimes there is a segfault in the next window
        self.main_window.deleteLater()
            
    def _iloop(self):
        self.current_state = name = target = None  # for except statement
        try:
            log('Initializing:', self.current_test)
            self.init_test()
            yield
            log('Running:', self.current_test)
            self.check_initial()
            assert self.reader.initial is not None
            yield from self.step(self.reader.initial, True)
            if self.current_state not in self.testdata.stateinfos:
                log_error('unknown state:', self.current_state)
                stateinfo = self.testdata.add_stateinfo(self.current_state)
                if not stateinfo.islimit:
                    log_error('state not in limit:', self.current_state)
            while True:
                name = self.testdata.get_random_transition(self.current_state)
                transition = self.testdata.transitions[name]
                yield from self.step(transition)
        except Quit as e:
            self.check_result()
            if self.write_mode == 'yes' or logger.result.errors and self.write_mode == 'error':
                self.reader.write_test()
                log('test data file written')
            log(e, self.current_test)
        except Exception:
            self.check_result()
            logf('exception in {}: {} -> {}', name,
                    self.current_state and self.current_state.tostr(),
                    target and target.tostr())
            sys.excepthook(*sys.exc_info())
        finally:
            if self.running:
                self.running = False
                self.main_window.close()
        
    @classmethod
    def wrap(cls, testdata_dir, tests, test_args):
        showtime = 'notime' not in test_args
        with Result(showtime) as result:
            temp_dir = tempfile.mkdtemp(prefix='pybiktest-')
            results = []
            cnt_tests = 0
            for test in tests:
                if '=' in test:
                    test, repeat_cnt = test.split('=', 1)
                    repeat_cnt = int(repeat_cnt)
                else:
                    repeat_cnt = 1
                with Result(showtime) as cnt_result:
                    results.append((test, cnt_result))
                    cnt_tests += repeat_cnt
                    for repeat_idx in range(repeat_cnt):
                        with Result(showtime) as logger.result:
                            if repeat_cnt > 1:
                                logf('Test: {} {}/{}', test, repeat_idx+1, repeat_cnt)
                            else:
                                log('Test:', test)
                            instance = cls(testdata_dir, test, test_args)
                            settings.reset()
                            instance.settings_file = tempfile.mktemp(prefix='settings.conf-', dir=temp_dir)
                            instance.games_file = tempfile.mktemp(prefix='games-', dir=temp_dir)
                            yield instance
                            if instance.running:
                                log_error('Unexpected end of testrunner')
                        log('Result:', test, logger.result)
                        log('')
                        os.remove(instance.settings_file)
                        del instance.settings_file
                        os.remove(instance.games_file)
                        del instance.games_file
                        result += logger.result
                        cnt_result.add(logger.result)
            with suppress(OSError):
                os.rmdir(temp_dir)
                
        for test, testresult in results:
            fmt = 'Test {0}: {1} errors' if testresult.errors else 'Test {0}: success'
            if showtime:
                fmt += ', {2}'
                dtime = testresult.fmt_dtime()
                logf(fmt, test, testresult.errors, dtime)
            else:
                logf(fmt, test, testresult.errors)
        if cnt_tests > 1:
            log('')
            logf('Summary ({}): {}', cnt_tests, result)
        
    



# -*- coding: utf-8 -*-

#  Copyright © 2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.

from math import sqrt

from pybiklib.debug import DEBUG_INVISIBLE
from pybiklib.modelcommon import epsilon
from .geom import Vector, Polyhedron, Plane, distance_axis_vert, distance_axis_edge, Face


N_ = lambda t: t


class ModelBase:
    bevel_width = 0.1
    
    def __init__(self):
        self.facekeys = [fn.replace(' ', '_') for fn in self.facenames]
        self.cell = self.create_single_cell()
        self.axes = self.create_axes()
        self.normals = {hf.face.id: hf.normal() for hf in self.cell.halffaces}
        
    def create_axes(self):
        hf_by_id = {hf.face.id: hf for hf in self.cell.halffaces}
        return [hf_by_id[s].center().normalised() for s in self.symbols]
        
    def scale_cell(self, polys, sizes):
        saxes = (a*s for a, s in zip(self.axes, sizes))
        scale1 = (sqrt(sum(f*f for f in a)) for a in zip(*self.axes))
        scaleS = (sqrt(sum(f*f for f in a)) for a in zip(*saxes))
        scale = tuple(s/s1 for s, s1 in zip(scaleS, scale1))
        polys.scale(scale)
        
    def split_cell(self, polys, sizes):
        slice_centers = []
        cuts = {}
        for iaxis, (axis, size) in enumerate(zip(self.axes, sizes)):
            slice_centers_axis = []
            axis = Vector(axis)
            plane = Plane(axis)
            split_low, split_high = self.splits(iaxis, size, [0, size])
            for plane.distance in self.splits(iaxis, size, range(1, size)):
                slice_centers_axis.append((split_low+plane.distance)/2)
                split_low = plane.distance
                split_id = polys.split_plane(plane, iaxis)
                cuts[split_id] = axis
            slice_centers_axis.append((split_low+split_high)/2)
            if split_low < split_high:
                slice_centers_axis.reverse()
            slice_centers.append(slice_centers_axis)
        return cuts, slice_centers
        
    @staticmethod
    def mark_invisible_faces(polys, cuts):
        for split_id, axis in cuts.items():
            visible_limit = 10000
            for cell in polys.cells:
                for hf in cell.halffaces:
                    if hf.face.type == 'face':
                        for he in hf.halfedges:
                            other_face = he.other.halfface.face
                            if other_face.type == 'cut' and other_face.id == split_id:
                                dist_ae = distance_axis_edge(axis, he.verts)
                                if dist_ae < visible_limit:
                                    visible_limit = dist_ae
            # mark all faces whose verts are within visible_limit around axis
            for face in polys.faces:
                if face.type == 'cut' and face.id == split_id:
                    for vert in face.verts:
                        dav = distance_axis_vert(axis, vert)
                        if dav > visible_limit - epsilon:
                            break
                    else:
                        face.type += '_removed'
                        
    @classmethod
    def remove_cells(self, polys, cuts):
        self.mark_invisible_faces(polys, cuts)
        if DEBUG_INVISIBLE:
            return
        polys.remove_cells(lambda c: all(f.type.endswith('_removed') for f in c.faces))
        
    def create_cells(self, sizes):
        self.cell.indices = (0,) * len(sizes)
        polys = Polyhedron()
        polys.cells = [self.cell]
        polys = polys.clone()
        self.scale_cell(polys, sizes)
        cuts, slice_centers = self.split_cell(polys, sizes)
        self.remove_cells(polys, cuts)
        return polys, slice_centers
        
        
class BrickModel (ModelBase):
    type = 'Brick'
    fileclass = 'b'
    name = N_('Brick')
    mformat = N_('{0}×{1}×{2}-Brick')
    sizenames = (N_('Width:'), N_('Height:'), N_('Depth:'))
    defaultsize = (4, 2, 3)
    
    symmetries = 2, 2, 2
    symbols = 'LDB'
    symbolsI = 'RUF'
    faces = 'UDLRFB'
    facenames = (N_('Up'), N_('Down'), N_('Left'), N_('Right'), N_('Front'), N_('Back'))
    default_rotation = (-30.,39.)
    
    @classmethod
    def norm_sizes(cls, size):
        y, z, x = sorted(size)
        return (x, y, z), (x, y, z)
        
    @staticmethod
    def create_single_cell():
        #              standard
        #             orientation    face symbols
        # *------*                     +---+
        # |\     |\      y             | U |
        # | *------*     |         +---+---+---+---+
        # | |    | |     o--x      | L | F | R | B |
        # *-|----* |      \        +---+---+---+---+
        #  \|     \|       z           | D |
        #   *------*                   +---+
        up = Face.polygon((-1., -1.), 4, ids='FRBL')
        cell = up.extrude_y(1., -1., ids='UD')
        return cell
        
    def splits(self, iaxis, size, isplits):
        return ((size - 2.*i) for i in isplits)
        
        
class TowerModel (BrickModel):
    type = 'Tower'
    name = N_('Tower')
    mformat = N_('{0}×{1}-Tower')
    sizenames = (N_('Basis:'), N_('Height:'))
    defaultsize = (2, 3)
    symmetries = 2, 4, 2
    
    @classmethod
    def norm_sizes(cls, size):
        x, y = size
        return (x, y), (x, y, x)
        
        
class CubeModel (BrickModel):
    type = 'Cube'
    name = N_('Cube')
    mformat = N_('{0}×{0}×{0}-Cube')
    sizenames = (N_('Size:'),)
    defaultsize = (3,)
    symmetries = 4, 4, 4
    
    @classmethod
    def norm_sizes(cls, size):
        x, = size
        return (x,), (x, x, x)
        
        
class VoidCubeModel (BrickModel):
    type = 'Cube3Void'
    name = N_('Void Cube')
    mformat = N_('Void Cube')
    sizenames = ()
    defaultsize = ()
    symmetries = 4, 4, 4
    
    @classmethod
    def norm_sizes(cls, unused_size):
        return (), (3, 3, 3)
        
    @classmethod
    def remove_cells(self, polys, cuts):
        polys.remove_cells(lambda c: sum(int(i==1) for i in c.indices) >= 2)
        
        
class TetrahedronModel (ModelBase):
    type = 'Tetrahedron'
    fileclass = 't'
    name = N_('Tetrahedron')
    mformat = N_('{0}-Tetrahedron')
    sizenames = (N_('Size:'),)
    defaultsize = (3,)
    
    symmetries = 3, 3, 3, 3
    symbols = 'LDBR'
    symbolsI = 'WXYZ'
    faces = 'DLRB'
    facenames = (N_('Down'), N_('Left'), N_('Right'), N_('Back'))
    default_rotation = (-10.,39.)
    
    @classmethod
    def norm_sizes(cls, size):
        x, = size
        return (x,), (x, x, x, x)
        
    _redge = sqrt(3/2)
    _itriangle = sqrt(2) / 2
    _rtriangle = sqrt(2)
    _itetrahedron = 1 / 2
    _rtetrahedron = 3 / 2
    
    _verts = [Vector((-_redge, -_itetrahedron, -_itriangle)),
              Vector(( _redge, -_itetrahedron, -_itriangle)),
              Vector(( 0.,      _rtetrahedron, 0.)),
              Vector(( 0.,     -_itetrahedron, _rtriangle))]
              
    @classmethod
    def create_single_cell(self):
        #  vertex      standard
        #  indices    orientation    face symbols
        # 2
        # |              y
        # |              |             +-----+
        # |              o--x        /L|B\ R/
        # 0------1        \        +---+---+
        #  \               z           |D/
        #   3                          +
        down = Face.polygon((0., -self._rtriangle), 3, ids='LBR')
        cell = down.pyramid(1.5, -.5, id='D')
        return cell
        
    def splits(self, iaxis, size, isplits):
        return ((i*2 - self._rtetrahedron*size) for i in isplits)
        
        
class TriangularPrism (ModelBase):
    type = 'Prism3'
    fileclass = 't'
    name = N_('Triangular Prism')
    mformat = N_('{0}×{1} Triangular Prism')
    sizenames = (N_('Basis:'), N_('Height:'))
    defaultsize = (3, 2)
    
    symmetries = 2, 3, 2, 2
    symbols = 'LDBR'
    symbolsI = 'XUYZ'
    faces = 'UDLRB'
    facenames = (N_('Up'), N_('Down'), N_('Left'), N_('Right'), N_('Back'))
    
    default_rotation = (-4.,39.)
    
    @classmethod
    def norm_sizes(cls, size):
        b, h = size
        if b > 6 or h > 6:
            return None, None
        return (b, h), (b, h, b, b)
        
    _rtriangle = sqrt(2)
    _itriangle = _rtriangle / 2
    
    @classmethod
    def create_single_cell(self):
        #              standard      polygon
        #             orientation  orientation
        # *------*
        # |\    /|       y          y
        # |  *   |       |          |
        # |  |   |       o--x       |
        # *--|---*        \         o-----x
        #  \ |  /          z
        #    *
        up = Face.polygon((0., -self._rtriangle), 3, ids='RBL')
        cell = up.extrude_y(1., -1., ids='UD')
        return cell
        
    split_arg = 0
    
    def splits(self, iaxis, size, isplits):
        if iaxis == 1:
            return ((i*2. - size) for i in isplits)
        else:
            low = self._rtriangle*size
            d = (self._itriangle + self._rtriangle) * (size/(size+self.split_arg))
            return ((i*d - low) for i in isplits)
            
    
class TriangularPrismComplex (TriangularPrism):
    type = 'Prism3Complex'
    fileclass = 't'
    name = N_('Triangular Prism (complex)')
    mformat = N_('{0}×{1} Triangular Prism (complex)')
    sizenames = (N_('Basis:'), N_('Height:'))
    defaultsize = (3, 2)
    
    split_arg = -1/3
    
    
class PentagonalPrism (ModelBase):
    type = 'Prism5'
    fileclass = 'p'
    name = N_('Pentagonal Prism')
    mformat = N_('{0}×{1} Pentagonal Prism')
    sizenames = (N_('Basis:'), N_('Height:'))
    defaultsize = (2, 2)
    
    symmetries = 2, 5, 2, 2, 2, 2
    symbols =  'AUBCDE'
    symbolsI = 'GLHIJK'
    faces = 'ULABCDE'
    facenames = (N_('Up'), N_('Down'),
                 N_('Front Right'), N_('Right'), N_('Back'), N_('Left'), N_('Front Left'))
    
    default_rotation = (-4.,39.)
    
    @classmethod
    def norm_sizes(cls, size):
        b, h = size
        if b > 6 or h > 6:
            return None, None
        return (b, h), (b, h, b, b, b, b)
        
    def create_single_cell(self):
        up = Face.polygon(Vector((0, -1.4)), 5, ids='ABCDE')
        self._ipentagon = up.halffaces[0].halfedge.center().length()
        upverts = iter(up.halffaces[0].verts)
        v1, unused, v3  = next(upverts), next(upverts), next(upverts)
        d3pentagon = (v1.point + v3.point).length() / 2
        self._hpentagon = (self._ipentagon - d3pentagon)/2
        
        cell = up.extrude_y(1., -1., ids='UL')
        return cell
        
    side_center = 0
        
    def splits(self, iaxis, size, isplits):
        if iaxis == 1:
            return ((i*2. - size) for i in isplits)
        else:
            dpentagon = self._ipentagon - self._hpentagon
            d = dpentagon*size/max(size-1+self.side_center, 1)
            low = self._hpentagon*size - (1-self.side_center)*d
            # values for i==0 is the same as for i==size.
            # its on the wrong side of the prism, but then most faces
            # have only 4 directions in quadrant mode
            return ((i*d + low if i>0 else size*self._ipentagon) for i in isplits)
            
        
class PentagonalPrismM (PentagonalPrism):
    type = 'Prism5M'
    fileclass = 'p'
    name = N_('Pentagonal Prism (stretched)')
    mformat = N_('{0}×{1} Pentagonal Prism (stretched)')
    
    side_center = 0.2
            
        
modeldefs = [CubeModel, TowerModel, BrickModel, TetrahedronModel, VoidCubeModel,
             TriangularPrism, TriangularPrismComplex,
             PentagonalPrism, PentagonalPrismM,
             ]
             




# pybiklib/gl.pxd
# generated with: ./tools/create-gl-pxd.py gl pybiklib/gl.pxd pybiklib/glarea.py pybiklib/gldraw.py pybiklib/gl.pxd

from libc.stddef cimport ptrdiff_t
from libc.stdint cimport int32_t, intptr_t, int8_t, uint8_t


# from /usr/include/GL/gl.h:

cdef extern from 'GL/gl.h':
    enum: GL_FALSE
    enum: GL_TRUE
    enum: GL_UNSIGNED_BYTE
    enum: GL_FLOAT
    enum: GL_POINTS
    enum: GL_LINES
    enum: GL_TRIANGLES
    enum: GL_CCW
    enum: GL_BACK
    enum: GL_CULL_FACE
    enum: GL_DEPTH_TEST
    enum: GL_RGB
    enum: GL_RGBA
    enum: GL_TEXTURE_2D
    enum: GL_VENDOR
    enum: GL_RENDERER
    enum: GL_VERSION
    enum: GL_EXTENSIONS
    enum: GL_DEPTH_BUFFER_BIT
    enum: GL_COLOR_BUFFER_BIT
    enum: GL_MULTISAMPLE
    enum: GL_SAMPLE_ALPHA_TO_COVERAGE
    enum: GL_SAMPLE_ALPHA_TO_ONE
    enum: GL_SAMPLE_COVERAGE
    enum: GL_SAMPLE_BUFFERS
    enum: GL_SAMPLES
    enum: GL_SAMPLE_COVERAGE_VALUE
    enum: GL_SAMPLE_COVERAGE_INVERT


# from /usr/include/GL/glext.h:

cdef extern from 'GL/glext.h':
    enum: GL_ARRAY_BUFFER
    enum: GL_STATIC_DRAW
    enum: GL_MAX_VERTEX_ATTRIBS
    enum: GL_FRAGMENT_SHADER
    enum: GL_VERTEX_SHADER
    enum: GL_DELETE_STATUS
    enum: GL_COMPILE_STATUS
    enum: GL_LINK_STATUS
    enum: GL_VALIDATE_STATUS
    enum: GL_INFO_LOG_LENGTH
    enum: GL_ATTACHED_SHADERS
    enum: GL_ACTIVE_UNIFORMS
    enum: GL_ACTIVE_UNIFORM_MAX_LENGTH
    enum: GL_ACTIVE_ATTRIBUTES
    enum: GL_ACTIVE_ATTRIBUTE_MAX_LENGTH
    enum: GL_SHADING_LANGUAGE_VERSION


# from /usr/include/GL/gl.h:

ctypedef unsigned int	GLenum
ctypedef unsigned char	GLboolean
ctypedef unsigned int	GLbitfield
ctypedef void		GLvoid
ctypedef int		GLint
ctypedef unsigned char	GLubyte
ctypedef unsigned int	GLuint
ctypedef int		GLsizei
ctypedef float		GLfloat
ctypedef float		GLclampf


# from /usr/include/GL/glext.h:

ctypedef ptrdiff_t GLsizeiptr
ctypedef ptrdiff_t GLintptr
ctypedef char GLchar


cdef extern from *:
    ctypedef GLchar* const_GLchar_ptr "const GLchar*"


# from /usr/include/GL/gl.h with:

cdef extern from 'GL/gl.h':
    cdef void  glClearColor( GLclampf red, GLclampf green, GLclampf blue, GLclampf alpha )
    cdef void  glClear( GLbitfield mask )
    cdef void  glCullFace( GLenum mode )
    cdef void  glFrontFace( GLenum mode )
    cdef void  glEnable( GLenum cap )
    cdef void  glDisable( GLenum cap )
    cdef GLboolean  glIsEnabled( GLenum cap )
    cdef void  glGetBooleanv( GLenum pname, GLboolean *params )
    cdef void  glGetFloatv( GLenum pname, GLfloat *params )
    cdef void  glGetIntegerv( GLenum pname, GLint *params )
    cdef GLubyte *  glGetString( GLenum name )
    cdef void  glViewport( GLint x, GLint y,
                                    GLsizei width, GLsizei height )
    cdef void  glDrawArrays( GLenum mode, GLint first, GLsizei count )
    cdef void  glReadPixels( GLint x, GLint y,
                                    GLsizei width, GLsizei height,
                                    GLenum format, GLenum type,
                                    GLvoid *pixels )
    cdef void  glTexSubImage2D( GLenum target, GLint level,
                                       GLint xoffset, GLint yoffset,
                                       GLsizei width, GLsizei height,
                                       GLenum format, GLenum type,
                                       GLvoid *pixels )
    cdef void  glActiveTexture( GLenum texture )


# from /usr/include/GL/glext.h with:

cdef extern from 'GL/glext.h':
    cdef void  glBindBuffer (GLenum target, GLuint buffer)
    cdef void  glDeleteBuffers (GLsizei n, GLuint *buffers)
    cdef void  glGenBuffers (GLsizei n, GLuint *buffers)
    cdef void  glBufferData (GLenum target, GLsizeiptr size, void *data, GLenum usage)
    cdef void  glBufferSubData (GLenum target, GLintptr offset, GLsizeiptr size, void *data)
    cdef void  glAttachShader (GLuint program, GLuint shader)
    cdef void  glBindAttribLocation (GLuint program, GLuint index, GLchar *name)
    cdef void  glCompileShader (GLuint shader)
    cdef GLuint  glCreateProgram ()
    cdef GLuint  glCreateShader (GLenum type)
    cdef void  glDeleteProgram (GLuint program)
    cdef void  glDeleteShader (GLuint shader)
    cdef void  glDetachShader (GLuint program, GLuint shader)
    cdef void  glEnableVertexAttribArray (GLuint index)
    cdef void  glGetActiveAttrib (GLuint program, GLuint index, GLsizei bufSize, GLsizei *length, GLint *size, GLenum *type, GLchar *name)
    cdef void  glGetActiveUniform (GLuint program, GLuint index, GLsizei bufSize, GLsizei *length, GLint *size, GLenum *type, GLchar *name)
    cdef GLint  glGetAttribLocation (GLuint program, GLchar *name)
    cdef void  glGetProgramiv (GLuint program, GLenum pname, GLint *params)
    cdef void  glGetProgramInfoLog (GLuint program, GLsizei bufSize, GLsizei *length, GLchar *infoLog)
    cdef void  glGetShaderiv (GLuint shader, GLenum pname, GLint *params)
    cdef void  glGetShaderInfoLog (GLuint shader, GLsizei bufSize, GLsizei *length, GLchar *infoLog)
    cdef GLint  glGetUniformLocation (GLuint program, GLchar *name)
    cdef void  glLinkProgram (GLuint program)
    cdef void  glShaderSource (GLuint shader, GLsizei count, GLchar **string, GLint *length)
    cdef void  glUseProgram (GLuint program)
    cdef void  glUniform1i (GLint location, GLint v0)
    cdef void  glUniformMatrix4fv (GLint location, GLsizei count, GLboolean transpose, GLfloat *value)
    cdef void  glValidateProgram (GLuint program)
    cdef void  glVertexAttribPointer (GLuint index, GLint size, GLenum type, GLboolean normalized, GLsizei stride, void *pointer)

# GL version 2.0 needed




#!/usr/bin/python3
# -*- coding: utf-8 -*-

#  Copyright © 2012-2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import sys, os
sys.path.insert(0, '.')
import pickletools
from multiprocessing import Pool, cpu_count
from collections import defaultdict, OrderedDict
from bisect import bisect_left
from array import array
from itertools import chain
from math import sin, cos, pi

from pybiklib.debug import DEBUG, DEBUG_MSG, DEBUG_NOLABEL, DEBUG_NOBEVEL, DEBUG_INVISIBLE, DEBUG_INDEXONLY
from pybiklib.modelcommon import epsilon, filebyteorder, get_texcoords_range
# executable script needs absolute imports
from buildlib.geom import roundeps, Vector
from buildlib.modeldef import modeldefs


dumps = None

def make_dumps(reproducible):
    global dumps
    import io
    import pickle
    if not reproducible:
        dumps = pickle.dumps
        return
    try:
        class _Pickler (pickle._Pickler):
            def save_dict(self, obj):
                if self.bin:
                    self.write(pickle.EMPTY_DICT)
                else:   # proto 0 -- can't use EMPTY_DICT
                    self.write(pickle.MARK + pickle.DICT)
                self.memoize(obj)
                self._batch_setitems(sorted(obj.items()))
                
            pickle._Pickler.dispatch[dict] = save_dict
            
        def _dumps(obj, protocol=None, *, fix_imports=True):
            f = io.BytesIO()
            try:
                _Pickler(f, protocol, fix_imports=fix_imports).dump(obj)
                return f.getvalue()
            except Exception:
                print('warning: using undocumented interface in module pickle failed, models will not be reproducible:')
                sys.excepthook(*sys.exc_info())
                return pickle.dumps(obj, protocol, fix_imports=fix_imports)
        _dumps({1:2, 3:4})
    except Exception:
        print('warning: using undocumented interface in module pickle failed, models will not be reproducible:')
        sys.excepthook(*sys.exc_info())
        _dumps = pickle.dumps
    dumps = _dumps
        
def equal_vector_fuzzy(v1, v2):
    for v1k, v2k in zip(v1, v2):
        if abs(v1k - v2k) > epsilon:
            return False
    return True
    
    
class VectorData:
    __slots__ = 'vectors indices sorted cnts dups toindex'.split()
    
    def __init__(self, fast):
        self.vectors = []
        self.indices = []
        self.sorted = []
        self.cnts = 0
        self.dups = 0
        self.toindex = self.toindex_fast if fast else self.toindex_dedup
        
    def toindex_dedup(self, vector):
        vector = vector.rounded()
        vector = [(0. if v == 0. else v) for v in vector]  # 0.0 == -0.0
        self.cnts += 1
        si = bisect_left(self.sorted, vector)
        i = len(self.vectors)
        if si < i and self.sorted[si] == vector:
            self.dups += 1
            return self.indices[si]
        self.vectors.append(vector)
        self.sorted.insert(si, vector)
        self.indices.insert(si, i)
        return i
        
    def toindex_fast(self, vector):
        vector = vector.rounded()
        i = len(self.vectors)
        self.vectors.append(vector)
        return i
        
        
class ModelFactory:
    mtype_attributes = (
            'name', 'mformat', 'sizenames', 'defaultsize',
            'symmetries', 'axes', 'symbols', 'symbolsI', 'faces', 'facekeys',
            'normal_rotation_symbols', 'rotation_symbols', 'rotation_matrices',
            'face_permutations', 'default_rotation')
        
    def __init__(self, modeldef):
        self.modeldef = modeldef()
        self.create_rotations()
        
    def __getattr__(self, attrname):
        return getattr(self.modeldef, attrname)
        
    @staticmethod
    def _matrix_equal(m1, m2):
        for line1, line2 in zip(m1, m2):
            for value1, value2 in zip(line1, line2):
                if abs(value1 - value2) > epsilon:
                    return False
        return True
        
    @staticmethod
    def _mult_matrix_vector3(matrix, vector):
        return Vector(sum(matrix[i][k]*vector[k] for k in range(3)) for i in range(3))
        
    @staticmethod
    def _mult_matrix(matrix1, matrix2):
        return [[sum(matrix1[i][k]*matrix2[k][j] for k in range(4)) for j in range(4)] for i in range(4)]
        
    @staticmethod
    def _create_rotation(axis, angle):
        angle = angle / 180. * pi
        sa = sin(angle)
        ca = cos(angle)
        e_ca = 1 - ca
        n1 = axis[0]
        n2 = axis[1]
        n3 = axis[2]
        m = [
            [n1*n1*e_ca + ca,    n1*n2*e_ca - n3*sa, n1*n3*e_ca + n2*sa, 0.],
            [n2*n1*e_ca + n3*sa, n2*n2*e_ca + ca,    n2*n3*e_ca - n1*sa, 0.],
            [n3*n1*e_ca - n2*sa, n3*n2*e_ca + n1*sa, n3*n3*e_ca + ca,    0.],
            [0.,    0.,     0.,     1.],
        ]
        #XXX: try to keep the matrix clean
        for y, line in enumerate(m):
            for x, value in enumerate(line):
                if abs(value) < epsilon:
                    m[y][x] = 0.
        return m
        
    def _create_permutation(self, matrix):
        permutation = {}
        for sym1, symI1, axis1 in zip(self.symbols, self.symbolsI, self.axes):
            for sym2, symI2, axis2 in zip(self.symbols, self.symbolsI, self.axes):
                axisR = self._mult_matrix_vector3(matrix, axis1)
                if axisR.equalfuzzy(axis2):
                    permutation[sym2] = sym1
                    permutation[symI2] = symI1
                elif axisR.inversfuzzy(axis2):
                    permutation[symI2] = sym1
                    permutation[sym2] = symI1
        return permutation
        
    def create_rotations(self):
        prim = []
        for axis, sym, symI, symmetry in zip(self.axes, self.symbols, self.symbolsI, self.symmetries):
            angle = 360. / symmetry
            # for some models (towers, bricks) rotations are equal to the invers rotations,
            # but the symbols are different
            prim.append((sym, self._create_rotation(axis, angle)))
            prim.append((symI, self._create_rotation(axis, -angle)))
        self.normal_rotation_symbols = {'': ''}
        transform = [['', [[1.,0.,0.,0.],[0.,1.,0.,0.],[0.,0.,1.,0.],[0.,0.,0.,1.]]]]
        for sp, p in prim:
            transform.append([sp, p])
        for sm, m in transform:
            for sp, p in prim:
                n = self._mult_matrix(m, p)
                sn = sm + sp
                for st, t in transform:
                    if self._matrix_equal(t, n):
                        self.normal_rotation_symbols[sn] = st
                        break
                else:
                    self.normal_rotation_symbols[sn] = sn
                    transform.append([sn, n])
        self.rotation_symbols = [s for s, m in transform]
        self.rotation_matrices = [m for s, m in transform]
        self.face_permutations = {s: self._create_permutation(m) for s, m in transform}
        
    def get_rotated_position(self, cells):
        rotated_position = {}
        centers = [c.center() for c in cells]
        for b, center in enumerate(centers):
            for sym, rotation in zip(self.rotation_symbols, self.rotation_matrices):
                coords = self._mult_matrix_vector3(rotation, center)
                for p, center2 in enumerate(centers):
                    if equal_vector_fuzzy(center2, coords):
                        if sym not in rotated_position:
                            rotated_position[sym] = [0] * len(cells)
                        rotated_position[sym][p] = b
                        break
                else:
                    assert False, 'not a permutation'
        return rotated_position
        
    @staticmethod
    def mark_invisible_faces_bevel(polys):
        for cell in polys.cells:
            for other in ['cut', 'bevel']:
                for hf in cell.halffaces:
                    if hf.face.type != 'bevel':
                        continue
                    other_removed = False
                    for he in hf.halfedges:
                        face_ = he.other.halfface.face
                        if face_.type == 'face':
                            break
                        if face_.type == other:
                            break
                        if face_.type.endswith('_removed'):
                            other_removed = True
                    else:
                        if other_removed:
                            hf.face.type += '_removed'
                            
    def get_data(self, sizes, vectordata):
        polys, slice_centers = self.create_cells(sizes)
        
        normals_polys = OrderedDict()  # for reproducible build
        for cell in polys.cells:
            for hf in cell.halffaces:
                key = (cell.indices, hf.face.id)
                normals_polys[key] = self.normals[hf.face.id] if hf.face.type == 'face' else hf.normal()
        data_cells_visible_faces = [[f.id for f in cell.faces if f.type == 'face'] for cell in polys.cells]
        normals = [self.normals[sym] for sym in self.faces]
        polys_verts = [v.point for v in polys.verts]
        data_texranges_mosaic = [list(get_texcoords_range(polys_verts, normal)) for normal in normals]
        data_normals = [vectordata.toindex(normal) for normal in normals]
        data_rotated_position = self.get_rotated_position(polys.cells)
        data_cell_indices = [cell.indices for cell in polys.cells]
        data_pick_polygons = list(self.gl_pick_polygons(polys.cells, vectordata))
        
        # beveled cells
        normals_polys = {k: vectordata.toindex(n) for k, n in normals_polys.items()}
        polys_beveled = polys
        polys_beveled.bevel(self.bevel_width)
        self.mark_invisible_faces_bevel(polys_beveled)
        if DEBUG_INVISIBLE:
            for face in polys_beveled.faces:
                if face.type.endswith('_removed'):
                    face.type = face.type.rsplit('_', 1)[0]
                else:
                    face.type += '_removed'
        polys_beveled.remove_faces(lambda f: f.type.endswith('_removed'))
        
        data_block_polygons = [[list(self.gl_label_polygons(cell, vectordata)),
                                list(self.gl_black_polygons(cell, normals_polys, vectordata))]
                                    for cell in polys_beveled.cells]
        return {
                'rotated_position': data_rotated_position,
                'cell_indices': data_cell_indices,
                'cells_visible_faces': data_cells_visible_faces,
                'texranges_mosaic': data_texranges_mosaic,
                'normals': data_normals,
                'block_polygons': data_block_polygons,
                'pick_polygons': data_pick_polygons,
                'slice_centers': slice_centers,
            }
        
    def gl_label_polygons(self, cell, vectordata):
        if DEBUG:
            if DEBUG_NOLABEL:
                return
        for halfface in cell.halffaces:
            if halfface.face.type != 'face':
                continue
            symbol = halfface.face.id
            faceno = self.faces.index(symbol)
            cell_vertices = [vectordata.toindex(he.verts[0].point) for he in halfface.halfedges]
            cell_vertices.append(faceno)
            yield cell_vertices
            
    @staticmethod
    def gl_black_polygons(cell, inormals, vectordata):
        if DEBUG:
            if DEBUG_NOBEVEL:
                return
        for halfface in cell.halffaces:
            if halfface.face.type == 'face':
                continue
            vertices = [he.verts[0] for he in halfface.halfedges]
            if halfface.face.type == 'cut':
                inormal = inormals[(cell.indices, halfface.face.id)]
                vnit = ((vectordata.toindex(v.point), inormal) for v in vertices)
                yield list(chain(*vnit))
            elif halfface.face.type == 'bevel':
                vnit = ((vectordata.toindex(v.point), inormals[(cell.indices, v.id)])
                            for v in vertices)
                yield list(chain(*vnit))
                
    def gl_pick_polygons(self, cells, vectordata):
        for cellidx, cell in enumerate(cells):
            for halfface in cell.halffaces:
                if halfface.face.type != 'face':
                    continue
                symbol = halfface.face.id
                face = self.faces.index(symbol)
                visible_face_indices = [i for i, he in enumerate(halfface.halfedges)
                                          if he.other.halfface.face.type == 'face']
                edges_iverts = [vectordata.toindex(he.verts[0].point) for he in halfface.halfedges]
                if len(visible_face_indices) == 0:
                    picktype = 0
                elif len(visible_face_indices) == 1:
                    if len(edges_iverts) == 3:
                        picktype = 1
                    elif len(edges_iverts) == 4:
                        picktype = 2
                    elif len(edges_iverts) == 6:
                        picktype = 4
                    else:
                        assert False, len(edges_iverts)
                    i = visible_face_indices[0]
                    edges_iverts = edges_iverts[i:] + edges_iverts[:i]
                elif len(visible_face_indices) == 2:
                    assert 3 <= len(edges_iverts) <= 5, len(edges_iverts)
                    i = visible_face_indices[0]
                    if len(edges_iverts) < 5:
                        picktype = 3
                    else:
                        picktype = 5
                    if i+1 == visible_face_indices[1]:
                        edges_iverts = edges_iverts[i:] + edges_iverts[:i]
                    elif visible_face_indices == [0, len(edges_iverts)-1]:
                        i = visible_face_indices[1]
                        edges_iverts = edges_iverts[i:] + edges_iverts[:i]
                    else:
                        picktype = -1
                else:
                    picktype = -1
                yield [cellidx, [face, picktype], edges_iverts]
                
        

class Dedup:
    def __init__(self):
        self.dedup_data = defaultdict(list)
        self.cnt_dups = 0
        self.cnt_values = 0
        
    def _float(self, value):
        self.cnt_values += 1
        dddlt = self.dedup_data['f']
        value = roundeps(value)
        if value == 0.0:  # 0.0 == -0.0
            value = 0.0  # for reproducible build
        didx = bisect_left(dddlt, value)
        if didx < len(dddlt) and dddlt[didx] == value:
            self.cnt_dups += 1
            return dddlt[didx], True, 'f'
        else:
            dddlt.insert(didx, value)
            return value, False, 'f'
            
    def _recursion(self, value, iterable):
        dedup = True
        vtypes = []
        for k, v in iterable:
            v, d, t = self.dedup(v)
            dedup = dedup and d
            vtypes.append(t)
            value[k] = v
        return dedup, '('+''.join(vtypes)+')'
        
    def _replace(self, value, dedup, vtype):
        self.cnt_values += 1
        dddlt = self.dedup_data[vtype]
        try:
            return self._replace_bisect(dddlt, value, dedup, vtype)
        except TypeError:
            return self._replace_index(dddlt, value, dedup, vtype)
            
    def _replace_index(self, dddlt, value, dedup, vtype):
        if not dedup:
            #assert value not in dddlt
            dddlt.append(value)
            return value, False, vtype
        try:
            didx = dddlt.index(value)
        except ValueError:
            dddlt.append(value)
            return value, False, vtype
        else:
            dvalue = dddlt[didx]
            self.cnt_dups += 1
            return dvalue, True, vtype
        
    def _replace_bisect(self, dddlt, value, unused_dedup, vtype):
        didx = bisect_left(dddlt, value)
        try:
            dvalue = dddlt[didx]
        except IndexError:
            dddlt.append(value)
            return value, False, vtype
        if value == dvalue:
            self.cnt_dups += 1
            return dvalue, True, vtype
        else:
            dddlt.insert(didx, value)
            return value, False, vtype
            
    def dedup(self, value):
        if isinstance(value, dict):
            value = {(sys.intern(k) if type(k) is str else k):v for k,v in value.items()}
            dedup, vtype = self._recursion(value, value.items())
            return self._replace(value, dedup, 'd' + vtype)
        elif isinstance(value, list):
            dedup, vtype = self._recursion(value, enumerate(value))
            return self._replace(tuple(value), dedup, 'a' + vtype)
        elif isinstance(value, tuple):
            value = list(value)
            dedup, vtype = self._recursion(value, enumerate(value))
            return self._replace(tuple(value), dedup, 'a' + vtype)
        elif isinstance(value, float):
            return self._float(value)
        elif isinstance(value, int):
            return value, True, 'i'
        elif isinstance(value, str):
            return sys.intern(value), True, 's'
        elif isinstance(value, type(None)):
            return value, True, 'n'
        else:
            assert False, type(value)
            
def pool_functions(parallel, reproducible):
    pool = None
    if reproducible:
        print('warning: using undocumented interface in module pickle for reproducible build')
    try:
        if parallel > 1:
            pool = Pool(processes=parallel, initializer=make_dumps, initargs=[reproducible])
    except OSError as e:
        print('process pool not available ({}):'.format(e))
        print('  deactivating multiprocessing')
    sys.stdout.flush() # when Pool(…) fails this line is 
    if pool is not None:
        return pool.imap_unordered
    else:
        make_dumps(reproducible)
        return map
        
#XXX: speedup parallel builds, longer jobs first
fileorder = 'b111 b101 b000 p111111 b010 p101111 b100 b110 p010000 p000000 b011 t1111 b001 t0000 t1011 t0100'.split() 

def get_datafilename(fileclass, sizes, fast):
    if fast:
        return fileclass + ''.join(str(s-1) for s in sizes)
    else:
        sizes = fileclass + ''.join(str((s-1)%2) for s in sizes)
        try:
            priority = fileorder.index(sizes) + 1
        except ValueError:
            priority = 99
        return '{:02}'.format(priority) + sizes
        
def enum_modelfiles(fast, maxsize, Factory):
    if isinstance(maxsize, tuple):
        minsize, maxsize = maxsize
    else:
        minsize = 1
    maxsize += 1
    def tuples(maxlen, part=()):
        if maxlen == len(part):
            yield part
        else:
            for i in range(minsize, maxsize):
                yield from tuples(maxlen, part+(i,))
    for size in tuples(len(Factory.sizenames)):
        norm_size, sizes = Factory.norm_sizes(size)
        if sizes is None:
            continue
        filename = get_datafilename(Factory.fileclass, sizes, fast)
        yield filename, sizes, size, norm_size
            
def pool_enum_modelfiles(dirname, testfunc, fast, maxsize):
    ignored = []
    modelfiles = []
    for Factory in modeldefs:
        for filename, *unused in enum_modelfiles(fast, maxsize, Factory):
            filename = os.path.join(dirname, filename)
            if filename in modelfiles or filename in ignored:
                continue
            if testfunc(filename):
                modelfiles.append(filename)
            else:
                ignored.append(filename)
    if modelfiles:
        for filename in sorted(ignored):
            print('skipping', filename)
    return sorted(modelfiles)
    
def pool_create_modelfiledata(path, fast, maxsize):
    filename = os.path.basename(path)
    savedata = {}
    vectordata = VectorData(fast)
    for Factory in modeldefs:
        sizes_list = []
        factory = None
        for _filename, sizes, *unused in enum_modelfiles(fast, maxsize, Factory):
            if _filename != filename:
                continue
            if sizes in sizes_list:
                continue
            sizes_list.append(sizes)
            if factory is None:
                factory = ModelFactory(Factory)
            savedata.setdefault(factory.type, {})[sizes] = factory.get_data(sizes, vectordata)
    vectors = array('f', chain.from_iterable(vectordata.vectors))
    return savedata, vectors, vectordata.cnts, vectordata.dups
    
def format_seconds(seconds):
    if seconds > 60:
        return '{:.0f}m {:.2f}s'.format(*divmod(seconds, 60))
    else:
        return '{:.2f}s'.format(seconds)
    
def pool_create_modelfile(filename, fast, pickle_protocol, maxsize):
    import time
    seconds = time.process_time()
    
    savedata, vectors, cnt_vectors, dup_vectors = pool_create_modelfiledata(filename, fast, maxsize)
    vals = dups = ''
    if not fast:
        dedup = Dedup()
        savedata = dedup.dedup(savedata)[0]
        if DEBUG_MSG:
            vals = '\n  vals: %6s  vec: %6s' % (dedup.cnt_values, cnt_vectors)
            dups = '\n  dups: %6s       %6s' % (dedup.cnt_dups, dup_vectors)
    
    savedata = dumps(savedata, pickle_protocol)
    datasize = ['{:.1f} kb'.format(len(savedata) / 1000), '---']
    vectorssize = '{:.1f} kb'.format(len(vectors) * vectors.itemsize / 1000)
    if not fast:
        savedata = pickletools.optimize(savedata)
        datasize[1] = '{:.1f} kb'.format(len(savedata) / 1000)
    seconds = time.process_time() - seconds
    
    vlen = array('I', [len(vectors)])
    if sys.byteorder != filebyteorder:
        vlen.byteswap()
        vectors.byteswap()
    with open(filename, 'wb') as datafile:
        datafile.write(savedata)
        vlen.tofile(datafile)
        vectors.tofile(datafile)
    return filename, seconds, 'generated {}, ({}, {}), {}, {} vectors {}{}{}'.format(
                    filename,
                    datasize[0], datasize[1], format_seconds(seconds), len(vectors) / 3, vectorssize,
                    vals, dups)
    
def pool_create_indexdata(fast, maxsize):
    savedata_type = {}
    savedata_types = []
    savedata_size = {}
    savedata_sizes = {}
    savedata_facenames = []
    savedata = {'type': savedata_type, 'types': savedata_types,
                'normsize': savedata_size, 'sizes': savedata_sizes,
                'facenames': savedata_facenames,
               }
    facekeys = []
    for Factory in modeldefs:
        factory = ModelFactory(Factory)
        savedata_type[factory.type] = {attr: getattr(factory, attr) for attr in factory.mtype_attributes}
        savedata_types.append(factory.type)
        savedata_size_type = {}
        savedata_sizes_type = {}
        savedata_size[factory.type] = savedata_size_type
        savedata_sizes[factory.type] = savedata_sizes_type
        for facekey, facename in zip(factory.facekeys, factory.facenames):
            if facekey not in facekeys:
                facekeys.append(facekey)
                savedata_facenames.append((facekey, facename))
        for filename, sizes, size, norm_size in enum_modelfiles(fast, maxsize, Factory):
            savedata_size_type[size] = norm_size
            if norm_size in savedata_sizes_type:
                assert [sizes, filename] == savedata_sizes_type[norm_size]
            else:
                savedata_sizes_type[norm_size] = [sizes, filename]
        size_range = ((min(vals), max(vals)) for vals in zip(*savedata_size_type.keys()))
        defaultsize = savedata_type[factory.type]['defaultsize']
        defaultsize = tuple(max(smin, min(s, smax)) for s, (smin, smax) in zip(defaultsize, size_range))
        savedata_type[factory.type]['defaultsize'] = defaultsize
    return savedata
    
def pool_create_indexfile(filename, fast, pickle_protocol, maxsize):
    import time
    seconds = time.process_time()
    
    savedata = pool_create_indexdata(fast, maxsize)
    vals = dups = ''
    if not fast:
        dedup = Dedup()
        savedata = dedup.dedup(savedata)[0]
        if DEBUG_MSG:
            vals = '\n  vals: %6s' % dedup.cnt_values
            dups = '\n  dups: %6s' % dedup.cnt_dups
    savedata = dumps(savedata, pickle_protocol)
    datasize = ['{:.1f} kb'.format(len(savedata) / 1000), '---']
    if not fast:
        savedata = pickletools.optimize(savedata)
        datasize[1] = '{:.1f} kb'.format(len(savedata) / 1000)
    seconds = time.process_time() - seconds
    
    with open(filename, 'wb') as datafile:
        datafile.write(savedata)
    return filename, seconds, 'generated {}, ({}, {}), {}{}{}'.format(
                    filename,
                    datasize[0], datasize[1], format_seconds(seconds),
                    vals, dups)
    
def pool_run(args):
    func, *args = args
    return func(*args)
    
def get_indexfilename(dirname):
    return os.path.join(dirname, 'index')
    
def create_modeldata(dirname, testfunc=None, parallel=1, fast=False, pickle_protocol=-1, maxsize=10, reproducible=False):
    # prepare jobs
    if testfunc is None:
        testfunc = lambda arg: True
    if DEBUG_INDEXONLY:
        modelfiles = []
    else:
        modelfiles = pool_enum_modelfiles(dirname, testfunc, fast, maxsize)
    jobs = [(pool_create_modelfile, m, fast, pickle_protocol, maxsize) for m in modelfiles]
    indexfilename = get_indexfilename(dirname)
    jobs.append((pool_create_indexfile, indexfilename, fast, pickle_protocol, maxsize))
    # run jobs
    if parallel is True:
        parallel = cpu_count()
    if not parallel or parallel < 1:
        parallel = 1
    realparallel = min(parallel, len(jobs))
    print('using {} / {} processes'.format(realparallel, parallel))
    imap_model = pool_functions(realparallel, reproducible)
    result = []
    for filename, seconds, lines in imap_model(pool_run, jobs):
        result.append((filename, seconds))
        print(lines)
        sys.stdout.flush()
    if DEBUG_MSG and not fast:
        result.sort(key=lambda fs: fs[1], reverse=True)
        if result != sorted(result, key=lambda fs: fs[0]):
            print('warning: reorder jobs to optimize parallel build')
            for filename, seconds in result:
                print(' ', os.path.basename(filename), format_seconds(seconds))
    
    
if __name__ == '__main__':
    create_modeldata('data/models', maxsize=(1,6))
    




#-*- coding:utf-8 -*-

#  Copyright © 2009, 2011-2015  B. Clausius <barcc@gmx.de>
#
#  This program is free software: you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program.  If not, see <http://www.gnu.org/licenses/>.


import os
from ast import literal_eval
from io import StringIO

from PyQt5.QtCore import QObject, QTimer
from PyQt5.QtCore import pyqtSignal as Signal

try:
    from .debug import debug
except ImportError:
    debug = print
from . import config


current_settings_version = 2

class KeyStore (QObject):
    changed = Signal(str)
    error = Signal(str)
    
    def __init__(self):
        super().__init__()
        
        self.schema = None
        self.filename = None
        self.keystore = {}
        
    def clone_key(self, keyfrom, keyto):
        self.schema[keyto] = self.schema[keyfrom]
        if keyfrom in self.keystore and keyto not in self.keystore:
            self.keystore[keyto] = self.keystore[keyfrom]
            
    def get_default(self, key):
        return self.schema[key][0]
    def get_validator(self, key):
        return self.schema[key][1]
        
    def get_range(self, key):
        validator = self.get_validator(key)
        if not isinstance(validator, (tuple, list)):
            raise ValueError('{} is not a range'.format(key))
        return validator
    
    def get_value(self, key):
        try:
            return self.keystore[key]
        except KeyError:
            return self.get_default(key)
            
    def get_nick(self, key):
        value = self.get_value(key)
        validator = self.get_validator(key)
        if not isinstance(validator, list):
            raise ValueError('{} is not an enum'.format(key))
        return validator[value]
        
    def set_value(self, key, value):
        self.keystore[key] = value
        self.changed.emit(key)
        
    def set_nick(self, key, nick):
        validator = self.get_validator(key)
        if not isinstance(validator, list):
            raise ValueError('{} is not an enum'.format(key))
        value = validator.index(nick)
        return self.set_value(key, value)
        
    def del_value(self, key):
        try:
            del self.keystore[key]
        except KeyError:
            pass # already the default value
        self.changed.emit(key)
        
    def find_match(self, key):
        def match(key, pattern):
            key = key.split('.')
            pattern = pattern.split('.')
            for k, p in zip(key, pattern):
                if k != p != '*':
                    return False
            return True
        for pattern in self.schema.keys():
            if match(key, pattern):
                return pattern
        return None
        
    def read_settings(self, filename):
        from .schema import schema, deprecated
        self.schema = schema
        self.filename = filename
        if not self.filename:
            return
        dirname = os.path.dirname(self.filename)
        if dirname and not os.path.exists(dirname):
            os.makedirs(dirname)
            
        # read settings
        try:
            with open(self.filename, 'rt', encoding='utf-8') as settings_file:
                lines = settings_file.readlines()
        except FileNotFoundError:
            lines = []
        for line in lines:
            # parse the line, discard invalid keys
            try:
                key, strvalue = line.split('=', 1)
            except ValueError:
                continue
            key = key.strip()
            strvalue = strvalue.strip()
            pattern = self.find_match(key)
            if pattern is None:
                continue
            self.schema[key] = self.schema[pattern]
            try:
                value = literal_eval(strvalue)
            except (ValueError, SyntaxError):
                continue
                
            # translate enums and validate values
            validator = self.get_validator(key)
            if validator is deprecated:
                pass
            elif isinstance(validator, list):
                try:
                    value = validator.index(value)
                except ValueError:
                    continue
            elif isinstance(validator, tuple):
                if not validator[0] <= value <= validator[1]:
                    continue
            elif validator is not None:
                if not validator(value):
                    continue
                    
            self.keystore[key] = value
            
    def dump(self, file, all=False):    # pylint: disable=W0622
        keydict = self.schema if all else self.keystore
        for key in sorted(keydict.keys()):
            if '*' in key:
                continue
            try:
                # translate enums
                value = self.get_nick(key)
            except ValueError:
                value = self.get_value(key)
            print(key, '=', repr(value), file=file)
        
    def write_settings(self):
        if not self.filename:
            return
        buf = StringIO()
        self.dump(buf)
        try:
            with open(self.filename, 'wt', encoding='utf-8') as settings_file:
                settings_file.write(buf.getvalue())
        except OSError as e:
            error_message = _('Settings can not be written to file: '
                            '{error_message}').format(error_message=e)
            debug(error_message)
            self.error.emit(error_message)
            
        
class Values:
    def __init__(self, key):
        object.__setattr__(self, '_key', key)
        object.__setattr__(self, '_subkeys_nodes', [])
        object.__setattr__(self, '_subkeys_leaves', [])
        
    def clone(self, key):
        values = Values(key)
        for subkey in self._subkeys_nodes:
            values._add_attr(subkey, self[subkey].clone(key+subkey+'.'))
        for subkey in self._subkeys_leaves:
            settings.keystore.clone_key(self._key + subkey, key + subkey)
            values._add_key(subkey)
        return values
        
    def sync(self, glob=None):
        glob2 = self['*'] if '*' in self._subkeys_nodes else None
        for subkey in self._subkeys_nodes:
            if subkey == '*':   # glob2 == self[subkey]
                glob2.sync()
            else:
                self[subkey].sync(glob2)
            if glob is not None and subkey in glob._subkeys_nodes:
                self[subkey].sync(glob[subkey])
        if glob is not None:
            for subkey in glob._subkeys_nodes:
                if subkey not in self._subkeys_nodes:
                    self._add_attr(subkey, glob[subkey].clone(self._key+subkey+'.'))
            for subkey in glob._subkeys_leaves:
                if subkey not in self._subkeys_leaves:
                    settings.keystore.clone_key(glob._key + subkey, self._key + subkey)
                    self._add_key(subkey)
                    
    def __getattr__(self, key):
        if '.' in key:
            value = self
            subkeys = key.split('.')
            for subkey in subkeys:
                value = value[subkey]
            return value
        if key and key[0] == '_':
            return super().__getattribute__(key)
        if key.endswith('_nick'):
            fullkey = self._key + key[:-5]
            return settings.keystore.get_nick(fullkey)
        if key.endswith('_range'):
            fullkey = self._key + key[:-6]
            return settings.keystore.get_range(fullkey)
        if key in self._subkeys_leaves:
            fullkey = self._key + key
            return settings.keystore.get_value(fullkey)
        if key in self._subkeys_nodes:
            return super().__getattribute__(key)
        if '*' in self._subkeys_leaves:
            fullkey = self._key + key
            settings.keystore.clone_key(self._key+'*', fullkey)
            self._add_key(key)
            return settings.keystore.get_value(fullkey)
        if '*' in self._subkeys_nodes:
            value = super().__getattribute__('*')
            value = value.clone(self._key+key+'.')
            self._add_attr(key, value)
            return value
        raise AttributeError('Unknown attribute "%s" in %s' % (key, self._key))
            
    def __setattr__(self, key, value):
        if '.' in key:
            val = self
            *subkeys, key = key.split('.')
            for subkey in subkeys:
                val = val[subkey]
            return val.__setattr__(key, value)
        if key.endswith('_nick'):
            fullkey = self._key + key[:-5]
            func = settings.keystore.set_nick
        elif key in self._subkeys_leaves:
            fullkey = self._key + key
            func = settings.keystore.set_value
        elif '*' in self._subkeys_leaves:
            fullkey = self._key + key
            settings.keystore.clone_key(self._key+'*', fullkey)
            self._add_key(key)
            func = settings.keystore.set_value
        else:
            raise AttributeError('unknown attribute {!r} in {!r}'.format(key, self._key))
        func(fullkey, value)
        if not settings._write_timer.isActive():
            settings._write_timer.start()
            
    def __delattr__(self, key):
        if '.' in key:
            value = self
            *subkeys, key = key.split('.')
            for subkey in subkeys:
                value = value[subkey]
            return value.__delattr__(key)
        if key in self._subkeys_leaves:
            fullkey = self._key + key
            settings.keystore.del_value(fullkey)
            if not settings._write_timer.isActive():
                settings._write_timer.start()
        else:
            raise AttributeError('unknown attribute "%s"' % key)
            
    def __getitem__(self, key):
        if type(key) is int:
            key = str(key)
        elif type(key) is tuple:
            key = '_'.join(str(v) for v in key)
        return getattr(self, key)
        
    def __setitem__(self, key, value):
        return setattr(self, str(key), value)
        
    def _add_attr(self, key, value):
        assert key not in self._subkeys_nodes
        assert key not in self._subkeys_leaves
        self._subkeys_nodes.append(key)
        super().__setattr__(key, value)
        
    def _add_key(self, key):
        assert key not in self._subkeys_nodes
        assert key not in self._subkeys_leaves
        self._subkeys_leaves.append(key)
        
        
class Settings:
    keystore = KeyStore()
    
    def __init__(self, key=''):
        object.__setattr__(self, '_key', key)
        object.__setattr__(self, '_array_len', 0)
        object.__setattr__(self, '_values', Values(''))
        object.__setattr__(self, '_write_timer', QTimer())
        self._write_timer.setSingleShot(True)
        self._write_timer.setInterval(5000)
        self._write_timer.timeout.connect(self.on_write_timer_timeout)
        
    def reset(self):
        self._write_timer.stop()
        self._write_timer.timeout.disconnect(self.on_write_timer_timeout)
        self.keystore.keystore.clear()
        self.__init__()
        
    def load(self, filename):
        self.keystore.read_settings(filename)
        keys = list(self.keystore.schema.keys())
        
        for key in keys:
            subkeys = key.split('.')
            settings_parent = self._values
            for subkey in subkeys[:-1]:
                if subkey.isdigit():
                    subkey = str(int(subkey))
                try:
                    settings_child = getattr(settings_parent, subkey)
                except (KeyError, AttributeError):
                    settings_child = Values(settings_parent._key + subkey + '.')
                    settings_parent._add_attr(subkey, settings_child)
                assert isinstance(settings_child, Values), settings_child
                settings_parent = settings_child
            else:
                if subkeys[-1] not in settings_parent._subkeys_leaves:
                    settings_parent._add_key(subkeys[-1])
        self._values.sync()
        version = settings.version
        if version != current_settings_version:
            settings.version = current_settings_version
        return version
            
    def on_write_timer_timeout(self):
        self.keystore.write_settings()
        
    def disconnect(self):
        self.keystore.changed.disconnect()
        self.keystore.error.disconnect()
        self._write_timer.stop()
        
    def flush(self):
        self.keystore.write_settings()
        
    def __getattr__(self, key):
        return getattr(self._values, key)
        
    def __setattr__(self, key, value):
        setattr(self._values, key, value)
        
    def __delattr__(self, key):
        delattr(self._values, key)
        
    def __getitem__(self, key):
        return self._values[key]
        

settings = Settings()

